# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging
import os
import re
import warnings
from datetime import datetime, timezone

import dask
import numpy as np
import pint
from dateutil.parser import parse as dtparse

from rsciio._docstrings import (
    FILENAME_DOC,
    LAZY_DOC,
    RETURNS_DOC,
    SHOW_PROGRESSBAR_DOC,
    SIGNAL_DOC,
)
from rsciio.utils._array import sarray2dict
from rsciio.utils._context_manager import get_progress_bar_context_manager
from rsciio.utils._decorator import jit_ifnumba
from rsciio.utils._dictionary import DTBox

_logger = logging.getLogger(__name__)


TVIPS_RECORDER_GENERAL_HEADER = [
    ("size", "u4"),  # likely the size of generalheader in bytes
    ("version", "u4"),  # 1 or 2
    ("dimx", "u4"),  # image size width
    ("dimy", "u4"),  # image size height
    ("bitsperpixel", "u4"),  # 8 or 16
    ("offsetx", "u4"),  # generally 0
    ("offsety", "u4"),
    ("binx", "u4"),  # camera binning
    ("biny", "u4"),
    ("pixelsize", "u4"),  # physical pixel size in nm
    ("ht", "u4"),  # high tension, voltage
    ("magtotal", "u4"),  # magnification/camera length
    ("frameheaderbytes", "u4"),  # number of bytes per frame header
    ("dummy", "S204"),  # placeholder contains TVIPS TVIPS TVIPS...
]


TVIPS_RECORDER_FRAME_HEADER = [
    ("num", "u4"),  # tends to cycle
    ("timestamp", "u4"),  # seconds since 1.1.1970
    ("ms", "u4"),  # additional milliseconds to the timestamp
    ("LUTidx", "u4"),  # related to color, useless
    ("fcurrent", "f4"),  # usually 0 for all frames
    ("mag", "u4"),  # same for all frames, can be different from magtotal
    ("mode", "u4"),  # 1 -> image, 2 -> diffraction
    ("stagex", "f4"),
    ("stagey", "f4"),
    ("stagez", "f4"),
    ("stagea", "f4"),
    ("stageb", "f4"),
    ("rotidx", "u4"),  # encodes information about the scan
    ("temperature", "f4"),  # cycles between 0.0 and 9.0 with step 1.0
    ("objective", "f4"),  # kind of randomly between 0.0 and 1.0
    # TODO: sometimes scan positions may be present in header, may require more reverse engineering
]


def _guess_image_mode(signal):
    """
    Guess whether the dataset contains images (1) or diffraction patterns (2).
    If no decent guess can be made, None is returned.
    """
    # original pixel scale
    scale = signal["axes"][-2]["scale"]
    unit = signal["axes"][-2]["units"]
    mode = None
    try:
        from rsciio.utils._units import _UREG

        pixel_size = scale * _UREG(unit)
    except (AttributeError, pint.UndefinedUnitError):
        pass
    else:
        if pixel_size.is_compatible_with("m"):
            mode = 1
        elif pixel_size.is_compatible_with("1/m"):
            mode = 2
        else:
            pass
    return mode


def _get_main_header_from_signal(signal, version=2, frame_header_extra_bytes=0):
    dt = np.dtype(TVIPS_RECORDER_GENERAL_HEADER)
    header = np.zeros((1,), dtype=dt)
    header["size"] = dt.itemsize
    header["version"] = version
    # original pixel scale
    mode = _guess_image_mode(signal)
    axes = signal["axes"]
    scale = axes[0]["scale"]
    offsetx = axes[-1]["offset"]
    offsety = axes[-2]["offset"]
    unit = axes[-2]["units"]
    if mode == 1:
        to_unit = "nm"
    elif mode == 2:
        to_unit = "1/nm"
    else:
        to_unit = ""
    if to_unit:
        from rsciio.utils._units import _UREG

        scale = round((scale * _UREG(unit)).to(to_unit).magnitude)
        offsetx = round((offsetx * _UREG(unit)).to(to_unit).magnitude)
        offsety = round((offsety * _UREG(unit)).to(to_unit).magnitude)
    else:
        warnings.warn(
            "Image scale units could not be converted, saving axes scales as is.",
            UserWarning,
        )
    metadata = DTBox(signal["metadata"], box_dots=True)
    header["dimx"] = axes[-1]["size"]
    header["dimy"] = axes[-2]["size"]
    header["offsetx"] = offsetx
    header["offsety"] = offsety
    header["pixelsize"] = scale
    header["bitsperpixel"] = signal["data"].dtype.itemsize * 8
    header["binx"] = 1
    header["biny"] = 1
    dtf = np.dtype(TVIPS_RECORDER_FRAME_HEADER)
    header["frameheaderbytes"] = dtf.itemsize + frame_header_extra_bytes
    header["dummy"] = "HYPERSPY " * 22 + "HYPERS"
    header["ht"] = metadata.get("Acquisition_instrument.TEM.beam_energy", 0)
    cl = metadata.get("Acquisition_instrument.TEM.camera_length", 0)
    mag = metadata.get("Acquisition_instrument.TEM.magnification", 0)
    if cl != 0 and mag != 0:
        header["magtotal"] = 0
    elif cl != 0 and mode == 2:
        header["magtotal"] = cl
    elif mag != 0 and mode == 1:
        header["magtotal"] = mag
    else:
        header["magtotal"] = 0
    return header


def _get_frame_record_dtype_from_signal(signal, extra_bytes=0):
    fhdtype = TVIPS_RECORDER_FRAME_HEADER.copy()
    if extra_bytes > 0:
        fhdtype.append(("extra", bytes, extra_bytes))
    dimx = signal["axes"][-1]["size"]
    dimy = signal["axes"][-2]["size"]
    fhdtype.append(("data", signal["data"].dtype, (dimy, dimx)))
    dt = np.dtype(fhdtype)
    return dt


def _is_valid_first_tvips_file(filename):
    """Check if the provided first tvips file path is valid"""
    filpattern = re.compile(r".+\_([0-9]{3})\.(.*)")
    match = re.match(filpattern, filename)
    if match is not None:
        num, ext = match.groups()
        if ext.lower() != "tvips":
            raise ValueError(f"Invalid tvips file: extension {ext}, must be tvips")
        if int(num) != 0:
            raise ValueError("Can only read video sequences starting with part 000")
        return True
    else:
        raise ValueError("Could not recognize as a valid tvips file")


def _find_auto_scan_start_stop(rotidxs):
    """Find the start and stop index in a rotator index array"""
    diff = rotidxs[1:] - rotidxs[:-1]
    index = np.where(diff > 0)[0]
    if index.size == 0:
        return None, None
    else:
        startx = index[0]
        if rotidxs[startx] == 0:
            startx += 1
        return startx, index[-1] + 1


@jit_ifnumba()
def _guess_scan_index_grid(rotidx, start, stop):
    indxs = np.zeros(rotidx[stop], dtype=np.int64)
    rotidx = rotidx[start : stop + 1]
    inv = 0  # index of the value we fill in
    for i in range(rotidx.shape[0]):
        if rotidx[inv] != rotidx[i]:
            # when we encounter a new value, we fill in indices
            pos_start = rotidx[inv] - 1
            pos_end = rotidx[i] - 1
            stack = np.arange(inv, i)[: pos_end - pos_start]
            indxs[pos_start:pos_end] = stack[-1]
            indxs[pos_start : pos_start + stack.shape[0]] = stack
            inv = i
    # the last value we fill in at the end
    indxs[rotidx[inv] - 1 :] = inv
    return indxs + start


def file_reader(
    filename,
    lazy=False,
    scan_shape=None,
    scan_start_frame=0,
    winding_scan_axis=None,
    hysteresis=0,
    rechunking="auto",
):
    """
    Read TVIPS stream file for in-situ and 4D STEM data.

    Parameters
    ----------
    %s
    %s
    scan_shape : str or tuple of int or None, optional
        By default the data is loaded as an image stack (1 navigation axis).
        A tuple of integers can be provided to indicate the shape of the
        navigation axes. For example, ``(3, 4)`` will have 3 scan points in the y
        direction and 4 in the x direction.
        If it concerns a 4D-STEM dataset, the (..., ``scan_y``, ``scan_x)``
        dimension can be provided. ``"auto"`` can also be selected, in which case
        the ``rotidx`` information in the frame headers will be used to try to
        reconstruct the scan. Additional navigation axes must be prepended. Since
        this only works for square scan grids and is prone to failure, this option
        is not recommended.
    scan_start_frame : int, optional
        Index of the first frame of the dataset to consider. Mainly relevant for
        4D-STEM datasets.If ``scan_shape = "auto"`` this is ignored.
    winding_scan_axis : str, optional
        If the acquisition software collected data without beam flyback but with
        a winding "snake" scan, then every second scan row or column needs to be
        reversed to make sense of the data. This can be indicated with values
        ``"x"`` or ``"y"``, depending on whether winding happened along the
        primary or secondary axis. By default, flyback scan without winding
        is assumed with ``x`` the fast scan and ``y`` the slow scan direction.
    hysteresis : int, optional
        Only applicable if ``winding_scan_axis`` is not ``None``, as it is likely
        there is an overshoot of a few pixels (2-5) every second scan row. This
        parameter allows shifts every second row by the indicated number of scan
        points to align even and odd scan rows. Default is 0, no hysteresis.
    rechunking : bool, str, dict, Default="auto"
        Only relevant when using lazy loading. If set to False each tvips file is
        a single chunk. For a better experience, with the default setting of
        ``"auto"`` rechunking is performed such that the navigation axes
        are optimally chunked and the signal axes are not chunked.
        If set to anything else, e.g. a dictionary, the value will be passed to
        the chunks argument in dask.array.rechunk.

    %s
    """
    # check whether we start at the first tvips file
    _is_valid_first_tvips_file(filename)

    # get all other filenames in case they exist
    other_files = []
    basename = filename[:-9]  # last bit: 000.tvips
    file_index = 1
    _, ext = os.path.splitext(filename)
    while True:
        fn = basename + "{:03d}{}".format(file_index, ext)
        if not os.path.exists(fn):
            break
        other_files.append(fn)
        file_index += 1

    # parse the header from the first file
    with open(filename, "rb") as f:
        f.seek(0)
        # read the main header in file 0
        header = np.fromfile(f, dtype=TVIPS_RECORDER_GENERAL_HEADER, count=1)
        dtype = np.dtype(f"u{header['bitsperpixel'][0] // 8}")
        dimx = header["dimx"][0]
        dimy = header["dimy"][0]
        # the size of the frame header varies with version
        if header["version"][0] == 1:
            increment = 12  # pragma: no cover
        elif header["version"][0] == 2:
            increment = header["frameheaderbytes"][0]
        else:
            raise NotImplementedError(
                f"This version {header.version} is not yet supported"
                " in HyperSpy. Please report this as an issue at "
                "https://github.com/hyperspy/hyperspy/issues."
            )  # pragma: no cover
        frame_header_dt = np.dtype(TVIPS_RECORDER_FRAME_HEADER)
        # the record must consume less bytes than reported in the main header
        if increment < frame_header_dt.itemsize:
            raise ValueError(
                "The frame header record consumes more bytes than stated in the main header"
            )  # pragma: no cover
        # save metadata
        original_metadata = {"tvips_header": sarray2dict(header)}
        # create custom dtype for memmap padding the frame_header as required
        extra_bytes = increment - frame_header_dt.itemsize
        record_dtype = TVIPS_RECORDER_FRAME_HEADER.copy()
        if extra_bytes > 0:
            record_dtype.append(("extra", bytes, extra_bytes))
        record_dtype.append(("data", dtype, (dimy, dimx)))

    # memmap the data
    records_000 = np.memmap(
        filename, mode="r", dtype=record_dtype, offset=header["size"][0]
    )
    # the array data
    all_array_data = [records_000["data"]]
    # in case we also want the frame header metadata later
    metadata_keys = np.array(TVIPS_RECORDER_FRAME_HEADER)[:, 0]
    metadata_000 = records_000[metadata_keys]
    all_metadata = [metadata_000]
    # also load data from other files
    for i in other_files:
        # no offset on the other files
        records = np.memmap(i, mode="r", dtype=record_dtype)
        all_metadata.append(records[metadata_keys])
        all_array_data.append(records["data"])
    if lazy:
        import dask.array as da

        data_stack = da.concatenate(all_array_data, axis=0)
    else:
        data_stack = np.concatenate(all_array_data, axis=0)

    # extracting some units/scales/offsets of the DP's or images
    mode = all_metadata[0]["mode"][0]
    DPU = "1/nm" if mode == 2 else "nm"
    SDP = header["pixelsize"][0]
    offsetx = header["offsetx"][0]
    offsety = header["offsety"][0]
    # modify the data if there is scan information
    # we construct a 2D array of indices to slice the data_stack
    if scan_shape is not None:
        # try to deduce start and stop of the scan based on rotator index
        if scan_shape == "auto":
            record_idxs = np.concatenate([i["rotidx"] for i in all_metadata])
            scan_start_frame, scan_stop_frame = _find_auto_scan_start_stop(record_idxs)
            if scan_start_frame is None or scan_stop_frame is None:
                raise ValueError(
                    "Scan start and stop information could not be automatically "
                    "determined. Please supply a scan_shape and scan_start_frame."
                )  # pragma: no cover
            total_scan_frames = record_idxs[scan_stop_frame]  # last rotator
            scan_dim = int(np.sqrt(total_scan_frames))
            if not np.allclose(scan_dim, np.sqrt(total_scan_frames)):
                raise ValueError(
                    "Scan was not square, please supply a scan_shape and start_frame."
                )
            scan_shape = (scan_dim, scan_dim)
            # there may be discontinuities which must be filled up
            indices = _guess_scan_index_grid(
                record_idxs, scan_start_frame, scan_stop_frame
            ).reshape(scan_shape)
        # scan shape and start are provided
        else:
            total_scan_frames = np.prod(scan_shape)
            max_frame_index = np.prod(data_stack.shape[:-2])
            final_frame = scan_start_frame + total_scan_frames
            if final_frame > max_frame_index:
                raise ValueError(
                    f"Shape {scan_shape} requires image index {final_frame - 1} "
                    f"which is out of bounds. Final frame index: {max_frame_index - 1}."
                )
            indices = np.arange(scan_start_frame, final_frame).reshape(scan_shape)

        # with winding scan, every second column or row must be inverted
        # due to hysteresis there is also a predictable offset
        if winding_scan_axis is not None:
            if winding_scan_axis in ["x", 0]:
                indices[..., ::2, :] = indices[..., ::2, :][..., :, ::-1]
                indices[..., ::2, :] = np.roll(
                    indices[..., ::2, :], hysteresis, axis=-1
                )
            elif winding_scan_axis in ["y", 1]:
                indices[..., :, ::2] = indices[..., :, ::2][..., ::-1, :]
                indices[..., :, ::2] = np.roll(
                    indices[..., :, ::2], hysteresis, axis=-2
                )
            else:
                raise ValueError("Invalid winding scan axis")

        with dask.config.set(**{"array.slicing.split_large_chunks": True}):
            data_stack = data_stack[indices.ravel()]
        data_stack = data_stack.reshape(*indices.shape, dimy, dimx)
        units = (indices.ndim - 2) * [""] + ["nm", "nm", DPU, DPU]
        names = (indices.ndim - 2) * [""] + ["y", "x", "dy", "dx"]
        # no scale information stored in the scan!
        scales = (indices.ndim - 2) * [1] + [1, 1, SDP, SDP]
        offsets = (indices.ndim - 2) * [0] + [0, 0, offsety, offsetx]
        # Create the axis objects for each axis
        dim = data_stack.ndim
        axes = [
            {
                "size": data_stack.shape[i],
                "index_in_array": i,
                "name": names[i],
                "scale": scales[i],
                "offset": offsets[i],
                "units": units[i],
                "navigate": True if i < len(scan_shape) else False,
            }
            for i in range(dim)
        ]
    else:
        # we load as a regular image stack
        units = ["s", DPU, DPU]
        names = ["time", "dy", "dx"]
        times = np.concatenate([i["timestamp"] + i["ms"] / 1000 for i in all_metadata])
        timescale = 1 if times.shape[0] <= 0 else times[1] - times[0]
        scales = [timescale, SDP, SDP]
        offsets = [times[0], offsety, offsetx]
        # Create the axis objects for each axis
        dim = data_stack.ndim
        axes = [
            {
                "size": data_stack.shape[i],
                "index_in_array": i,
                "name": names[i],
                "scale": scales[i],
                "offset": offsets[i],
                "units": units[i],
                "navigate": True if i == 0 else False,
            }
            for i in range(dim)
        ]
    dtobj = datetime.fromtimestamp(all_metadata[0]["timestamp"][0])
    date = dtobj.date().isoformat()
    time = dtobj.time().isoformat()
    current = all_metadata[0]["fcurrent"][0]
    stagex = all_metadata[0]["stagex"][0]
    stagey = all_metadata[0]["stagey"][0]
    stagez = all_metadata[0]["stagez"][0]
    stagealpha = all_metadata[0]["stagea"][0]
    stagebeta = all_metadata[0]["stageb"][0]
    # mag = all_metadata[0]["mag"][0]  # TODO it is unclear what this value is
    focus = all_metadata[0]["objective"][0]
    metadata = {
        "General": {
            "original_filename": os.path.split(filename)[1],
            "date": date,
            "time": time,
            "time_zone": "UTC",
        },
        "Acquisition_instrument": {
            "TEM": {
                "magnification": header["magtotal"][0],
                "beam_energy": header["ht"][0],
                "beam_current": current,
                "defocus": focus,
                "Stage": {
                    "tilt_alpha": stagealpha,
                    "tilt_beta": stagebeta,
                    "x": stagex,
                    "y": stagey,
                    "z": stagez,
                },
            },
        },
    }

    if lazy:
        if rechunking:
            if rechunking == "auto":
                navdims = data_stack.ndim - 2
                chunks = {ax_index: "auto" for ax_index in range(navdims)}
                chunks[navdims] = None
                chunks[navdims + 1] = None
            else:
                chunks = rechunking
            data_stack = data_stack.rechunk(chunks)

    if mode == 2:
        metadata["Signal"] = {"signal_type": "diffraction"}
    # TODO at the moment hyperspy doesn't have a signal type for mode==1, imaging

    dictionary = {
        "data": data_stack,
        "axes": axes,
        "metadata": metadata,
        "original_metadata": original_metadata,
        "mapping": {},
    }

    return [
        dictionary,
    ]


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)


def file_writer(
    filename,
    signal,
    max_file_size=None,
    version=2,
    frame_header_extra_bytes=0,
    mode=None,
    show_progressbar=True,
):
    """
    Write signal to TVIPS file.

    Parameters
    ----------
    %s
    %s
    max_file_size : int or None, default=None
        Approximate maximum size of individual files in bytes. In this way a
        dataset can be split into multiple files. A file needs to be at least the
        size of the main header in the first file plus one frame and its frame
        header. Sequential files are denoted by a suffix _xxx starting from _000.
        By default (``None``) there is no maximum and the entire dataset is stored in a
        single file.
    version : int, default=2
        TVIPS file format version (only version ``1`` or ``2`` supported).
    frame_header_extra_bytes : int, default=0
        Number of bytes to pad the frame headers with.
    mode : int or None, default=None
        ``1`` for imaging, ``2`` for diffraction. By default, the mode is
        guessed from signal type and signal units.
    %s"""
    # only signal2d is allowed
    axes = signal["axes"]
    metadata = DTBox(signal["metadata"], box_dots=True)
    signal_dim = len([axis for axis in axes if not axis["navigate"]])
    nav_shape = [axis["size"] for axis in axes if axis["navigate"]]
    num_frames = np.prod(nav_shape) if nav_shape else 0
    if signal_dim != 2:
        raise ValueError("Only Signal2D supported for writing to TVIPS file.")
    fnb, ext = os.path.splitext(filename)
    if fnb.endswith("_000"):
        fnb = fnb[:-4]
    main_header = _get_main_header_from_signal(
        signal, version, frame_header_extra_bytes
    )
    # frame header + frame dtype
    record_dtype = _get_frame_record_dtype_from_signal(signal, frame_header_extra_bytes)
    total_file_size = main_header.itemsize + num_frames * record_dtype.itemsize
    if max_file_size is None:
        max_file_size = total_file_size
    minimum_file_size = main_header.itemsize + record_dtype.itemsize
    if max_file_size < minimum_file_size:
        warnings.warn(
            f"The minimum file size for this dataset is {minimum_file_size} bytes"
        )
        max_file_size = minimum_file_size
    # frame metadata
    start_date_str = metadata.get("General.date", "1970-01-01")
    start_time_str = metadata.get("General.time", "00:00:00")
    tz = metadata.get("General.time_zone", "UTC")
    datetime_str = f"{start_date_str} {start_time_str} {tz}"
    time_dt = dtparse(datetime_str)
    time_dt_utc = time_dt.astimezone(timezone.utc)
    # workaround for timestamp not working on Windows, see https://bugs.python.org/issue37527
    BEGIN = datetime(1970, 1, 1, 0).replace(tzinfo=timezone.utc)
    timestamp = (time_dt_utc - BEGIN).total_seconds()
    if num_frames:
        nav_units = signal["axes"][-3]["units"]
        nav_increment = signal["axes"][-3]["scale"]
        try:
            from rsciio.utils._units import _UREG

            time_increment = (nav_increment * _UREG(nav_units)).to("ms").magnitude
        except (AttributeError, pint.UndefinedUnitError, pint.DimensionalityError):
            time_increment = 1
    # imaging or diffraction
    if mode is None:
        mode = _guess_image_mode(signal)
    mode = 2 if mode is None else mode
    stagex = metadata.get("Acquisition_instrument.TEM.Stage.x", 0)
    stagey = metadata.get("Acquisition_instrument.TEM.Stage.y", 0)
    stagez = metadata.get("Acquisition_instrument.TEM.Stage.z", 0)
    stagea = metadata.get("Acquisition_instrument.TEM.tilt_alpha", 0)
    stageb = metadata.get("Acquisition_instrument.TEM.tilt_beta", 0)
    # TODO: is fcurrent actually beam current??
    fcurrent = metadata.get("Acquisition_instrument.TEM.beam_current", 0)
    frames_to_save = num_frames if num_frames else 1
    current_frame = 0
    file_index = 0
    data = signal["data"]
    if num_frames:
        fdata = data.reshape((num_frames, axes[-2]["size"], axes[-1]["size"]))
    while frames_to_save != 0:
        suffix = "_" + (f"{file_index}".zfill(3))
        filename = fnb + suffix + ext
        if file_index == 0:
            with open(filename, "wb") as f:
                main_header.tofile(f)
                file_location = f.tell()
                open_mode = "r+"
        else:
            file_location = 0
            open_mode = "w+"
        frames_saved = (max_file_size - file_location) // record_dtype.itemsize
        # last file can contain fewer images
        if frames_to_save < frames_saved:
            frames_saved = frames_to_save
        file_memmap = np.memmap(
            filename,
            dtype=record_dtype,
            mode=open_mode,
            offset=file_location,
            shape=frames_saved,
        )
        # fill in the metadata
        file_memmap["mode"] = mode
        file_memmap["stagex"] = stagex
        file_memmap["stagey"] = stagey
        file_memmap["stagez"] = stagez
        file_memmap["stagea"] = stagea
        file_memmap["stageb"] = stageb
        file_memmap["fcurrent"] = fcurrent
        rotator = np.arange(current_frame, current_frame + frames_saved)
        milliseconds = rotator * time_increment
        timestamps = (timestamp + milliseconds / 1000).astype(int)
        milliseconds = milliseconds % 1000
        file_memmap["timestamp"] = timestamps
        file_memmap["ms"] = milliseconds
        file_memmap["rotidx"] = rotator + 1
        data = fdata[current_frame : current_frame + frames_saved]
        if signal["attributes"]["_lazy"]:
            with get_progress_bar_context_manager(show_progressbar)():
                data.store(file_memmap["data"])
        else:
            file_memmap["data"] = data
        file_memmap.flush()
        file_index += 1
        frames_to_save -= frames_saved
        current_frame += frames_saved


file_writer.__doc__ %= (
    FILENAME_DOC.replace("read", "write to"),
    SIGNAL_DOC,
    SHOW_PROGRESSBAR_DOC,
)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging
import os
import warnings
from datetime import datetime as dt

import numpy as np

from rsciio import __version__
from rsciio._docstrings import (
    ENCODING_DOC,
    FILENAME_DOC,
    LAZY_UNSUPPORTED_DOC,
    RETURNS_DOC,
    SIGNAL_DOC,
)
from rsciio.utils._dictionary import DTBox

_logger = logging.getLogger(__name__)


# For a description of the EMSA/MSA format, including the meaning of the
# following keywords:
# https://www.amc.anl.gov/ANLSoftwareLibrary/02-MMSLib/XEDS/EMMFF/EMMFF.IBM/Emmff.Total

US_MONTHS_D2A = {
    "01": "JAN",
    "02": "FEB",
    "03": "MAR",
    "04": "APR",
    "05": "MAY",
    "06": "JUN",
    "07": "JUL",
    "08": "AUG",
    "09": "SEP",
    "10": "OCT",
    "11": "NOV",
    "12": "DEC",
}

US_MONTH_A2D = dict([reversed(i) for i in US_MONTHS_D2A.items()])

keywords = {
    # Required parameters
    "FORMAT": {"dtype": str, "mapped_to": None},
    "VERSION": {"dtype": str, "mapped_to": None},
    "TITLE": {"dtype": str, "mapped_to": "General.title"},
    "DATE": {"dtype": str, "mapped_to": None},
    "TIME": {"dtype": str, "mapped_to": None},
    "OWNER": {"dtype": str, "mapped_to": None},
    "NPOINTS": {"dtype": float, "mapped_to": None},
    "NCOLUMNS": {"dtype": float, "mapped_to": None},
    "DATATYPE": {"dtype": str, "mapped_to": None},
    "XPERCHAN": {"dtype": float, "mapped_to": None},
    "OFFSET": {"dtype": float, "mapped_to": None},
    # Optional parameters
    # Signal1D characteristics
    "SIGNALTYPE": {"dtype": str, "mapped_to": None},
    "XLABEL": {"dtype": str, "mapped_to": None},
    "YLABEL": {"dtype": str, "mapped_to": None},
    "XUNITS": {"dtype": str, "mapped_to": None},
    "YUNITS": {"dtype": str, "mapped_to": None},
    "CHOFFSET": {"dtype": float, "mapped_to": None},
    "COMMENT": {"dtype": str, "mapped_to": None},
    # Microscope
    "BEAMKV": {"dtype": float, "mapped_to": "Acquisition_instrument.TEM.beam_energy"},
    "EMISSION": {"dtype": float, "mapped_to": None},
    "PROBECUR": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.beam_current",
    },
    "BEAMDIAM": {"dtype": float, "mapped_to": None},
    "MAGCAM": {"dtype": float, "mapped_to": None},
    "OPERMODE": {"dtype": str, "mapped_to": None},
    "CONVANGLE": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.convergence_angle",
    },
    # Specimen
    "THICKNESS": {"dtype": float, "mapped_to": "Sample.thickness"},
    "XTILTSTGE": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Stage.tilt_alpha",
    },
    "YTILTSTGE": {"dtype": float, "mapped_to": None},
    "XPOSITION": {"dtype": float, "mapped_to": None},
    "YPOSITION": {"dtype": float, "mapped_to": None},
    "ZPOSITION": {"dtype": float, "mapped_to": None},
    # EELS
    # in ms:
    "INTEGTIME": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EELS.exposure",
    },
    # in ms:
    "DWELLTIME": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EELS.dwell_time",
    },
    "COLLANGLE": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EELS.collection_angle",
    },
    "ELSDET": {"dtype": str, "mapped_to": None},
    # EDS
    "ELEVANGLE": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EDS.elevation_angle",
    },
    "AZIMANGLE": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EDS.azimuth_angle",
    },
    "SOLIDANGLE": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EDS.solid_angle",
    },
    "LIVETIME": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EDS.live_time",
    },
    "REALTIME": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EDS.real_time",
    },
    "FWHMMNKA": {
        "dtype": float,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EDS."
        + "energy_resolution_MnKa",
    },
    "TBEWIND": {"dtype": float, "mapped_to": None},
    "TAUWIND": {"dtype": float, "mapped_to": None},
    "TDEADLYR": {"dtype": float, "mapped_to": None},
    "TACTLYR": {"dtype": float, "mapped_to": None},
    "TALWIND": {"dtype": float, "mapped_to": None},
    "TPYWIND": {"dtype": float, "mapped_to": None},
    "TBNWIND": {"dtype": float, "mapped_to": None},
    "TDIWIND": {"dtype": float, "mapped_to": None},
    "THCWIND": {"dtype": float, "mapped_to": None},
    "EDSDET": {
        "dtype": str,
        "mapped_to": "Acquisition_instrument.TEM.Detector.EDS.EDS_det",
    },
}


def parse_msa_string(string, filename=None):
    """
    Parse an EMSA/MSA file content.

    Parameters
    ----------
    string : str
        It must complain with the EMSA/MSA standard.
    filename : str, None
        The filename.

    Returns
    -------
    list
        List of a single dictionary to be returned by ``file_reader``.
    """
    if not hasattr(string, "readlines"):
        string = string.splitlines()
    parameters = {}
    mapped = DTBox(box_dots=True)
    y = []
    # Read the keywords
    data_section = False
    for line in string:
        if data_section is False:
            if line[0] == "#":
                try:
                    key, value = line.split(": ")
                    value = value.rstrip()
                except ValueError:
                    key = line
                    value = None
                key = key.strip("#").strip()

                # Support multiple lines in the TITLE and COMMENT fields,
                # which are the only fields that can be multiline according to the standard.
                if key in ["TITLE", "COMMENT"] and key in parameters.keys():
                    parameters[key] += value
                elif key != "SPECTRUM":
                    parameters[key] = value
                else:
                    data_section = True
        else:
            # Read the data
            if line[0] != "#" and line.strip():
                line_ = line.replace(",", " ").strip().split()
                if parameters["DATATYPE"] == "XY":
                    y.append(float(line_[1]))
                    if len(line_) > 2:
                        # NCOLUMNS is 2
                        y.append(float(line_[3]))
                elif parameters["DATATYPE"] == "Y":
                    data = [float(i) for i in line_]
                    y.extend(data)
    # We rewrite the format value to be sure that it complies with the
    # standard, because it will be used by the writer routine
    parameters["FORMAT"] = "EMSA/MAS Spectral Data File"

    # Convert the parameters to the right type and map some
    # TODO: the msa format seems to support specifying the units of some
    # parameters. We should add this feature here
    for parameter, value in parameters.items():
        # Some parameters names can contain the units information
        # e.g. #AZIMANGLE-dg: 90.
        if "-" in parameter:
            clean_par, units = parameter.split("-")
            clean_par, units = clean_par.strip(), units.strip()
        else:
            clean_par, units = parameter, None
        if value and clean_par in keywords:
            type_ = keywords[clean_par]["dtype"]
            try:
                parameters[parameter] = type_(value)
            except Exception:
                error = f"The {parameter} keyword value, {value} could \
                    not be converted to the right type."
                if "e" in value.lower():
                    # Normally, the offending misspelling is a space in the
                    # scientific notation, e.g. 2.0 E-06
                    try:
                        parameters[parameter] = type_(value.replace(" ", ""))
                    except Exception:  # pragma: no cover
                        _logger.exception(error)
                else:
                    # Some files have two values separated by a space
                    # https://eelsdb.eu/wp-content/uploads/2017/03/Cu4O3-O-K.msa
                    try:
                        parameters[parameter] = type_(value.split(" ")[0])
                    except Exception:  # pragma: no cover
                        _logger.exception(error)

            if keywords[clean_par]["mapped_to"] is not None:
                mapped.set_item(keywords[clean_par]["mapped_to"], parameters[parameter])
                if units is not None:
                    mapped.set_item(keywords[clean_par]["mapped_to"] + "_units", units)
    if "TIME" in parameters and parameters["TIME"]:
        try:
            if len(parameters["TIME"].split(":")) == 2:
                time = dt.strptime(parameters["TIME"], "%H:%M")
            elif len(parameters["TIME"].split(":")) == 3:
                # Some files may have seconds in the time field,
                # even though the standard does not include seconds
                time = dt.strptime(parameters["TIME"], "%H:%M:%S")
            mapped.set_item("General.time", time.time().isoformat())
        except ValueError as e:
            _logger.warning(
                "Possible malformed TIME field in msa file. The time "
                f"information could not be retrieved.: {e}"
            )

    malformed_date_error = "Possibly malformed DATE in msa file. The date information could not be retrieved."
    if "DATE" in parameters and parameters["DATE"]:
        try:
            day, month, year = parameters["DATE"].split("-")
            if month.upper() in US_MONTH_A2D:
                month = US_MONTH_A2D[month.upper()]
                date = dt.strptime("-".join((day, month, year)), "%d-%m-%Y")
                mapped.set_item("General.date", date.date().isoformat())
            else:
                _logger.warning(malformed_date_error)
        except (
            ValueError
        ) as e:  # Error raised if split does not return 3 elements in this case
            _logger.warning(malformed_date_error + ": %s" % e)

    axes = [
        {
            "size": len(y),
            "index_in_array": 0,
            "name": parameters["XLABEL"] if "XLABEL" in parameters else "",
            "scale": parameters["XPERCHAN"] if "XPERCHAN" in parameters else 1,
            "offset": parameters["OFFSET"] if "OFFSET" in parameters else 0,
            "units": parameters["XUNITS"] if "XUNITS" in parameters else "",
            "navigate": False,
        }
    ]
    if filename is not None:
        mapped.set_item("General.original_filename", os.path.split(filename)[1])

    # Defaults to empty signal type, which is Signal1D for HyperSpy
    signal_type = ""
    if "SIGNALTYPE" in parameters and parameters["SIGNALTYPE"]:
        if parameters["SIGNALTYPE"] == "ELS":
            signal_type = "EELS"
        elif parameters["SIGNALTYPE"] == "CLS":
            signal_type = "CL"
        else:
            if parameters["SIGNALTYPE"] not in [
                "EDS",
                "WDS",
                "AES",
                "PES",
                "XRF",
                "GAM",
            ]:  # pragma: no cover
                warnings.warn(
                    f"SIGNALTYPE {parameters['SIGNALTYPE']} does not correspond "
                    "to any of the valid strings according to the MSA file "
                    "format definition."
                )
        if parameters["SIGNALTYPE"] == "EDS":
            # get the beam energy key
            beam_energy = None
            for key in parameters.keys():
                if key.startswith("BEAMKV"):
                    beam_energy = float(parameters[key])
                    break
            signal_type = "EDS"
            if beam_energy is not None and beam_energy > 30:
                # The MSA format does not specify the microscope type,
                # but EDS spectra acquired above 30 kV will most likely follow
                # TEM type analysis methods
                signal_type += "_TEM"
            else:
                signal_type += "_SEM"

    mapped.set_item("Signal.signal_type", signal_type)
    if "YUNITS" in parameters.keys():
        yunits = f"({parameters['YUNITS']})"
    else:
        yunits = ""
    if "YLABEL" in parameters.keys():
        quantity = f"{parameters['YLABEL']}"
    else:
        if mapped.Signal.signal_type == "EELS":
            quantity = "Electrons"
            if not yunits:
                yunits = "(Counts)"
        elif "EDS" in mapped.Signal.signal_type:
            quantity = "X-rays"
            if not yunits:
                yunits = "(Counts)"
        else:
            quantity = ""
    if quantity or yunits:
        quantity_units = f"{quantity} {yunits}"
        mapped.set_item("Signal.quantity", quantity_units.strip())

    dictionary = {
        "data": np.array(y),
        "axes": axes,
        "metadata": mapped.to_dict(),
        "original_metadata": parameters,
    }
    file_data_list = [
        dictionary,
    ]
    return file_data_list


def file_reader(filename, lazy=False, encoding="latin-1"):
    """
    Read an MSA file.

    Parameters
    ----------
    %s
    %s
    %s

    %s
    """
    if lazy is not False:
        raise NotImplementedError("Lazy loading is not supported.")

    with open(filename, encoding=encoding, errors="replace") as spectrum_file:
        return parse_msa_string(string=spectrum_file, filename=filename)


file_reader.__doc__ %= (FILENAME_DOC, LAZY_UNSUPPORTED_DOC, ENCODING_DOC, RETURNS_DOC)


def file_writer(filename, signal, format="Y", separator=", ", encoding="latin-1"):
    """
    Write signal to an MSA file.

    Parameters
    ----------
    %s
    %s
    format : str, default="Y"
        Specify whether the X-axis (energy/wavelength) should also be saved with
        the data. The default, ``"Y"`` omits the X-axis in the file. The alternative,
        ``"XY"``, saves the calibrated signal axis as first column.
    separator : str, Default=", "
        Change the column separator. However, if a different separator is chosen
        the resulting file will not comply with the MSA/EMSA standard and
        RosettaSciIO and other software may not be able to read it.
    %s

    Examples
    --------
    >>> from rsciio.msa import file_writer
    >>> file_writer("file.msa", signal, encoding="utf8")
    """
    loc_kwds = {}
    FORMAT = "EMSA/MAS Spectral Data File"
    md = DTBox(signal["metadata"], box_dots=True)
    if signal["original_metadata"].get("FORMAT", None) == FORMAT:
        loc_kwds = signal["original_metadata"]
        if format is not None:
            loc_kwds["DATATYPE"] = format
        else:
            if "DATATYPE" in loc_kwds:
                format = loc_kwds["DATATYPE"]
    else:
        if format is None:
            format = "Y"
        if "General.date" in md:
            date = dt.strptime(md.General.date, "%Y-%m-%d")
            date_str = date.strftime("%d-%m-%Y")
            day, month, year = date_str.split("-")
            month = US_MONTHS_D2A[month]
            loc_kwds["DATE"] = "-".join((day, month, year))
        if "General.item" in md:
            time = dt.strptime(md.General.time, "%H:%M:%S")
            loc_kwds["TIME"] = time.strftime("%H:%M")
    if md.Signal.signal_type in ["EDS_SEM", "EDS_TEM"]:
        loc_kwds["SIGNALTYPE"] = "EDS"
    elif md.Signal.signal_type in ["EELS"]:
        loc_kwds["SIGNALTYPE"] = "ELS"
    elif md.Signal.signal_type in ["CL", "CL_SEM", "CL_STEM"]:
        loc_kwds["SIGNALTYPE"] = "CLS"
    elif md.Signal.signal_type not in [
        "EDS",
        "WDS",
        "ELS",
        "AES",
        "PES",
        "XRF",
        "CLS",
        "GAM",
    ]:
        loc_kwds["SIGNALTYPE"] = ""
    else:
        loc_kwds["SIGNALTYPE"] = md.Signal.signal_type
    keys_from_signal = {
        # Required parameters
        "FORMAT": FORMAT,
        "VERSION": "1.0",
        "TITLE": getattr(md, "General.title", ""),
        "DATE": "",
        "TIME": "",
        "OWNER": "",
        "NPOINTS": signal["axes"][0]["size"],
        "NCOLUMNS": 1,
        "DATATYPE": format,
        "SIGNALTYPE": "",
        "XPERCHAN": signal["axes"][0]["scale"],
        "OFFSET": signal["axes"][0]["offset"],
        # Signal1D characteristics
        "XLABEL": signal["axes"][0]["name"],
        # "YLABEL" : getattr(md, "Signal.quantity", ""),
        "XUNITS": signal["axes"][0]["units"],
        #        'YUNITS' : '',
    }
    if "COMMENT" not in loc_kwds:
        loc_kwds["COMMENT"] = f"File created by RosettaSciIO version {__version__}"
        # Microscope
        #        'BEAMKV' : ,
        #        'EMISSION' : ,
        #        'PROBECUR' : ,
        #        'BEAMDIAM' : ,
        #        'MAGCAM' : ,
        #        'OPERMODE' : ,
        #        'CONVANGLE' : ,
        # Specimen
        #        'THICKNESS' : ,
        #        'XTILTSTGE' : ,
        #        'YTILTSTGE' : ,
        #        'XPOSITION' : ,
        #        'YPOSITION' : ,
        #        'ZPOSITION' : ,
        #
        # EELS
        # 'INTEGTIME' : , # in ms
        # 'DWELLTIME' : , # in ms
        #        'COLLANGLE' : ,
        #        'ELSDET' :  ,

    # Update the loc_kwds with the information retrieved from the signal class
    for key, value in keys_from_signal.items():
        if key not in loc_kwds or value != "":
            loc_kwds[key] = value

    for key, dic in keywords.items():
        if dic["mapped_to"] is not None:
            if "SEM" in md.Signal.signal_type:
                dic["mapped_to"] = dic["mapped_to"].replace("TEM", "SEM")
            if dic["mapped_to"] in md:
                loc_kwds[key] = eval("md.%s" % dic["mapped_to"])

    with open(filename, "w", encoding=encoding, errors="ignore") as f:
        # Remove the following keys from loc_kwds if they are in
        # (although they shouldn't)
        for key in ["SPECTRUM", "ENDOFDATA"]:
            if key in loc_kwds:
                del loc_kwds[key]

        f.write("#%-12s: %s\u000d\u000a" % ("FORMAT", loc_kwds.pop("FORMAT")))
        f.write("#%-12s: %s\u000d\u000a" % ("VERSION", loc_kwds.pop("VERSION")))
        for keyword, value in loc_kwds.items():
            if (
                isinstance(value, str)
                and keyword in ["TITLE", "COMMENT"]
                and len(value) > 0
            ):
                # Split the value into multiple lines if it is too long, and write each line with the same keyword
                for i in range(0, len(value), 64):
                    f.write("#%-12s: %s\u000d\u000a" % (keyword, value[i : i + 64]))
            else:
                f.write("#%-12s: %s\u000d\u000a" % (keyword, value))

        f.write("#%-12s: Spectral Data Starts Here\u000d\u000a" % "SPECTRUM")

        if format == "XY":
            axis_dict = signal["axes"][0]
            axis = axis_dict["offset"] + axis_dict["scale"] * np.arange(
                axis_dict["size"]
            )
            for x, y in zip(axis, signal["data"]):
                f.write("%g%s%g" % (x, separator, y))
                f.write("\u000d\u000a")
        elif format == "Y":
            for y in signal["data"]:
                f.write("%f%s" % (y, separator))
                f.write("\u000d\u000a")
        else:
            raise ValueError("format must be one of: None, 'XY' or 'Y'")

        f.write("#%-12s: End Of Data and File" % "ENDOFDATA")


file_writer.__doc__ %= (
    FILENAME_DOC.replace("read", "write to"),
    SIGNAL_DOC,
    ENCODING_DOC.replace("read", "write"),
)



#!/usr/bin/env python3
import os
import subprocess


def main():
    os.chdir(r"rsciio/tests")
    cmd = ("python", "registry_utils.py")

    return subprocess.call(cmd)


if __name__ == "__main__":
    # Script used by the pre-commit hook
    exit(main())



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

"""
EventList reconstruction for Tofwerk TofDAQ HDF5 files.

This module contains the peak-integration reconstruction functions that convert
raw TDC EventList data into a 4-D ``(depth, y, x, m/z)`` peak-counts array,
replicating the processing performed by the Tofwerk proprietary software.

Public API
----------
compute_peak_data_from_eventlist
    Reconstruct the 4-D peak-integrated array from a raw EventList.
"""

import logging
import re
from typing import Any

import numpy as np

from rsciio.utils._decorator import jit_ifnumba

_logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------
# PeakTable helpers
# ---------------------------------------------------------------------------


def _peak_table_from_list(
    peak_table_list: list[dict[str, Any]],
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Convert a peak table list of dicts back to arrays suitable for integration.

    Parameters
    ----------
    peak_table_list : list of dict
        As returned by :func:`_peak_table_to_list`.

    Returns
    -------
    masses : numpy.ndarray, float64
    peak_low : numpy.ndarray, float64
    peak_high : numpy.ndarray, float64
    """
    masses = np.array([p["mass"] for p in peak_table_list], dtype=float)
    peak_low = np.array(
        [p["lower_integration_limit"] for p in peak_table_list], dtype=float
    )
    peak_high = np.array(
        [p["upper_integration_limit"] for p in peak_table_list], dtype=float
    )
    return masses, peak_low, peak_high


# ---------------------------------------------------------------------------
# EventList / peak data reconstruction helpers
# ---------------------------------------------------------------------------


def count_active_channels(ini: str) -> int:
    """
    Return the number of active TDC recording channels from the INI config string.

    TofDAQ records events from multiple ADC/TDC channels simultaneously
    (e.g. Ch1 and Ch3 in a typical fibTOF setup).  Each active channel
    contributes an independent copy of every ion event to the EventList,
    so the raw event count per pixel is multiplied by this factor.

    Parameters
    ----------
    ini : str
        Decoded text from the HDF5 root attribute
        ``file.attrs["Configuration File Contents"]``. This is the TofDAQ
        configuration stored as an INI-style string, for example containing
        entries such as ``Ch1Record=1`` and ``Ch2Record=0``.

    Returns
    -------
    int
        Number of active channels (at least 1).

    Examples
    --------
    Read the embedded TofDAQ configuration from a file and count the active
    recording channels:

    >>> ini = _decode(file.attrs["Configuration File Contents"])
    >>> count_active_channels(ini)
    2

    A minimal INI-style example:

    >>> count_active_channels("[TOFParameter]\\nCh1Record=1\\nCh2Record=0\\nCh3Record=1")
    2
    """
    matches = re.findall(r"Ch\dRecord=(\d)", ini)
    return max(sum(int(m) for m in matches), 1)


def _flatten_event_row(row: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Pack all events from one row of pixels into a flat contiguous buffer.

    Parameters
    ----------
    row : numpy.ndarray
        1-D object array of shape ``(nx,)`` for one (depth, y) row, as
        returned by ``el[w][s]`` or ``el[w, s, :]``.

    Returns
    -------
    flat : numpy.ndarray, dtype int64
        All events concatenated in pixel order (x varies).
    offsets : numpy.ndarray, shape (nx + 1,), dtype int64
        ``flat[offsets[x]:offsets[x+1]]`` are the events for pixel ``x``.
    """
    _nx = len(row)
    lengths = np.array([len(row[x]) for x in range(_nx)], dtype=np.int64)
    offsets = np.empty(_nx + 1, dtype=np.int64)
    offsets[0] = 0
    np.cumsum(lengths, out=offsets[1:])
    flat = (
        np.concatenate(list(row)).astype(np.int64)
        if offsets[-1] > 0
        else np.empty(0, dtype=np.int64)
    )
    return flat, offsets


@jit_ifnumba(cache=True)
def _accumulate_peak_counts(
    flat: np.ndarray,
    offsets: np.ndarray,
    nx: int,
    clock_ratio: int,
    nbr_samples: int,
    mass_axis: np.ndarray,
    sorted_low: np.ndarray,
    sorted_high: np.ndarray,
    npeaks: int,
    result_sx: np.ndarray,
) -> None:  # pragma: no cover
    """
    Accumulate per-pixel peak counts for one row into *result_sx*.

    Parameters
    ----------
    flat : int64 array
        Flat event buffer from :func:`_flatten_event_row`.
    offsets : int64 array, shape (nx + 1,)
        Pixel start/end indices into *flat*.
    nx : int
        Number of x pixels in the row.
    clock_ratio : int
        TDC-to-ADC sample index divisor.
    nbr_samples : int
        Length of MassAxis; events outside [0, nbr_samples) are discarded.
    mass_axis : float64 array
        Calibrated mass (Da) for each ADC sample index.
    sorted_low, sorted_high : float64 arrays, shape (npeaks,)
        Integration window bounds in ascending mass order.
    npeaks : int
        Number of peaks.
    result_sx : float32 array, shape (nx, npeaks)
        Output accumulator, pre-zeroed by caller.  Indexed by sorted peak order.
    """
    for x in range(nx):
        start = offsets[x]
        end = offsets[x + 1]
        for i in range(start, end):
            sample = flat[i] // clock_ratio
            if sample < 0 or sample >= nbr_samples:
                continue
            mass = mass_axis[sample]
            bin_idx = np.searchsorted(sorted_low, mass, side="right") - 1
            if bin_idx >= 0 and mass <= sorted_high[bin_idx]:
                result_sx[x, bin_idx] += 1


def compute_peak_data_from_eventlist(
    event_list: np.ndarray,
    mass_axis: np.ndarray,
    nbr_samples: int,
    clock_ratio: int,
    normalization: int,
    peak_table: list[dict[str, Any]],
    show_progressbar: bool = True,
    dtype: np.dtype = np.uint16,
) -> np.ndarray:
    """
    Reconstruct the 4-D peak-integrated array from raw EventList data.

    This replicates the processing performed by the Tofwerk proprietary
    software when opening a raw file:

    1. For each pixel's variable-length EventList (raw TDC timestamps),
       convert timestamps to ADC sample indices by integer division with the
       TDC-to-ADC clock ratio (``SampleInterval / ClockPeriod``, typically 64).
    2. Look up the calibrated mass for each sample index via ``MassAxis``.
    3. Count events that fall within each peak's integration window.
    4. Divide by ``NbrWaveforms × NActiveChannels`` — the number of times each
       physical ion is recorded in the EventList (once per ToF cycle per active
       recording channel).

    .. note::
        If ``numba`` is installed, each depth slice is read in a single h5py
        call (instead of one call per row), then each row is packed into a flat
        int64 buffer and processed by a JIT-compiled kernel.  Otherwise falls
        back to a vectorised NumPy path that allocates per-pixel arrays but is
        still significantly faster than a pure-Python triple loop.

    Parameters
    ----------
    event_list : array-like
        Ragged object array of shape ``(nwrites, nsegs, nx)`` of uint16 TDC
        timestamps, one variable-length array per pixel.  Accepts h5py
        variable-length datasets or numpy object arrays (as loaded by
        ``signal="event_list"``).
    mass_axis : numpy.ndarray
        1-D array of shape ``(nbr_samples,)`` with calibrated mass (Da) for
        each ADC sample index, from ``FullSpectra/MassAxis``.
    nbr_samples : int
        Length of ``mass_axis``; events outside ``[0, nbr_samples)`` are
        discarded.
    clock_ratio : int
        TDC-to-ADC sample index divisor (``SampleInterval / ClockPeriod``).
        Pass 1 if ``ClockPeriod`` is not available.
    normalization : int
        ``NbrWaveforms × NActiveChannels`` — the divisor used to convert raw
        event counts to per-waveform ion counts.
    peak_table : list of dict
        Integration windows.  Each dict must have keys
        ``lower_integration_limit`` and ``upper_integration_limit`` (Da).
    show_progressbar : bool, default=True
        Whether to show tqdm progress bars during reconstruction.
    dtype : numpy.dtype, default=np.uint16
        Output array dtype.  Counts are accumulated as float32 internally (for
        the normalization division) and cast to this dtype on return.

    Returns
    -------
    numpy.ndarray
        Shape ``(nwrites, nsegs, nx, npeaks)``, cast to ``dtype``.
        Identical in structure to ``PeakData/PeakData`` in a pre-processed file.
    """
    el = event_list
    mass_axis = np.asarray(mass_axis, dtype=np.float64)
    _, peak_low, peak_high = _peak_table_from_list(peak_table)
    npeaks = len(peak_table)
    nwrites, nsegs, nx = el.shape

    # Sort peak windows for O(log n) assignment via searchsorted
    sort_idx = np.argsort(peak_low)
    sorted_low = peak_low[sort_idx]
    sorted_high = peak_high[sort_idx]
    # Inverse permutation: maps sorted-peak index back to original peak order.
    # Used to avoid numpy's advanced-indexing-on-LHS shape quirk where
    # result[w, s, :, sort_idx] = arr yields shape (npeaks, nx) not (nx, npeaks).
    inv_sort_idx = np.argsort(sort_idx)

    result = np.zeros((nwrites, nsegs, nx, npeaks), dtype=np.float32)

    _logger.info(
        f"Reconstructing peak data from EventList "
        f"({nwrites} depth slices × {nsegs} × {nx} pixels × {npeaks} peaks)."
    )

    # Prefer the numba JIT path when numba is installed: events are packed into a
    # flat contiguous buffer per row and processed in a single compiled loop,
    # avoiding per-pixel Python object overhead entirely.  Falls back to a
    # vectorised NumPy path (one array allocation per pixel) when numba is absent.
    try:
        import numba  # noqa: F401

        _use_numba = True
    except ImportError:
        _use_numba = False

    try:
        from tqdm.auto import tqdm as _tqdm
    except ImportError:
        _tqdm = None  # type: ignore[assignment]

    def _make_pbar(n, desc="Reconstructing peak data"):
        if _tqdm is not None and show_progressbar:
            return _tqdm(total=n, desc=desc, unit=" slices")
        return None

    if _use_numba:
        # numba path: read one depth slice per h5py call, then process one row
        # at a time -- flatten the row into a contiguous int64 buffer and
        # dispatch the JIT-compiled kernel per row.
        result_sx = np.zeros((nx, npeaks), dtype=np.float32)
        pbar = _make_pbar(nwrites)
        for w in range(nwrites):
            slice_data = el[w]  # one h5py read covers all nsegs rows
            for s in range(nsegs):
                flat, offsets = _flatten_event_row(slice_data[s])
                result_sx[:] = 0
                _accumulate_peak_counts(
                    flat,
                    offsets,
                    nx,
                    clock_ratio,
                    nbr_samples,
                    mass_axis,
                    sorted_low,
                    sorted_high,
                    npeaks,
                    result_sx,
                )
                result[w, s] = result_sx[:, inv_sort_idx]
            if pbar is not None:
                pbar.update(1)
        if pbar is not None:
            pbar.close()
    else:
        # NumPy fallback path: read one depth slice at a time, then iterate
        # over rows and pixels explicitly, converting each pixel's
        # variable-length event array to masses and binning with searchsorted
        # + bincount.  Slower than the numba path but avoids any compiled
        # dependency.
        pbar = _make_pbar(nwrites)
        for w in range(nwrites):
            slice_data = el[w]  # one read covers all nsegs rows for this depth slice
            for s in range(nsegs):
                row = slice_data[s]
                for x in range(nx):
                    events = row[x]
                    if len(events) == 0:
                        continue
                    processed = events // clock_ratio
                    valid = (processed >= 0) & (processed < nbr_samples)
                    if not valid.any():
                        continue
                    event_masses = mass_axis[processed[valid]]
                    bin_idx = (
                        np.searchsorted(sorted_low, event_masses, side="right") - 1
                    )
                    in_peak = (
                        (bin_idx >= 0)
                        & (bin_idx < npeaks)
                        & (event_masses <= sorted_high[bin_idx])
                    )
                    counts = np.bincount(bin_idx[in_peak], minlength=npeaks)
                    result[w, s, x, sort_idx] = counts
            if pbar is not None:
                pbar.update(1)
        if pbar is not None:
            pbar.close()

    result /= normalization
    return result.astype(dtype)



# -*- coding: utf-8 -*-
#
# Copyright 2016 Petras Jokubauskas
# Copyright 2016 The HyperSpy developers
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any project and source this library is coupled.
# If not, see <https://www.gnu.org/licenses/#GPL>.
#
# This python library subset provides read functionality of
#  Bruker bcf files.
# The basic reading capabilities of proprietary AidAim Software(tm)
#  SFS (Single File System) (used in bcf technology) is present in
#  the same library.

import codecs
import io
import logging
import xml.etree.ElementTree as ET
from ast import literal_eval
from datetime import datetime
from math import ceil
from os.path import basename, splitext
from struct import unpack as strct_unp
from zlib import decompress as unzip_block

import dask
import numpy as np

from rsciio._docstrings import FILENAME_DOC, LAZY_DOC, RETURNS_DOC
from rsciio.utils._date_time import msfiletime_to_unix
from rsciio.utils.xml import XmlToDict, sanitize_msxml_float

_logger = logging.getLogger(__name__)

warn_once = True

try:
    from rsciio.bruker import unbcf_fast

    fast_unbcf = True
    _logger.info("The fast cython based bcf unpacking library were found")
except ImportError:  # pragma: no cover
    fast_unbcf = False
    _logger.info(
        """unbcf_fast library is not present...
Falling back to slow python only backend."""
    )

# create dictionizer customized to Bruker Xml streams:
x2d = XmlToDict(dub_attr_pre_str="XmlClass", tags_to_flatten="ClassInstance")


class Container:
    pass


class SFSTreeItem:
    """Class to manage one internal sfs file.

    Reading, reading in chunks, reading and extracting, reading without
    extracting even if compression is present.

    Attributes:
    item_raw_string -- the bytes from sfs file table describing the file
    parent -- the item higher hierarchicaly in the sfs file tree

    Methods:
    read_piece, setup_compression_metadata, get_iter_and_properties,
    get_as_BytesIO_string
    """

    def __init__(self, item_raw_string, parent):
        self.sfs = parent
        (
            self._pointer_to_pointer_table,
            self.size,
            create_time,
            mod_time,
            some_time,
            self.permissions,
            self.parent,
            _,
            self.is_dir,
            _,
            name,
            _,
        ) = strct_unp("<iQQQQIi176s?3s256s32s", item_raw_string)
        self.create_time = msfiletime_to_unix(create_time)
        self.mod_time = msfiletime_to_unix(mod_time)
        self.some_time = msfiletime_to_unix(some_time)
        self.name = name.strip(b"\x00").decode("utf-8")
        self.size_in_chunks = self._calc_pointer_table_size()
        if self.is_dir == 0:
            self._fill_pointer_table()

    def _calc_pointer_table_size(self):
        n_chunks = ceil(self.size / self.sfs.usable_chunk)
        return n_chunks

    def _fill_pointer_table(self):
        """Parse the sfs and populate self.pointers table.

        self.pointer is the sfs pointer table containing addresses of
        every chunk of the file.

        The pointer table if the file is big can extend through many
        sfs chunks. Differently than files, the pointer table of file have no
        table of pointers to the chunks. Instead if pointer table is larger
        than sfs chunk, the chunk header contains next chunk number (address
        can be calculated using known chunk size and global offset) with
        continuation of file pointer table, thus it have to be read and filled
        consecutive.
        """
        # table size in number of chunks:
        n_of_chunks = ceil(self.size_in_chunks / (self.sfs.usable_chunk // 4))
        with open(self.sfs.filename, "rb") as fn:
            if n_of_chunks > 1:
                next_chunk = self._pointer_to_pointer_table
                temp_string = io.BytesIO()
                for dummy1 in range(n_of_chunks):
                    fn.seek(self.sfs.chunksize * next_chunk + 0x118)
                    next_chunk = strct_unp("<I", fn.read(4))[0]
                    fn.seek(28, 1)
                    temp_string.write(fn.read(self.sfs.usable_chunk))
                temp_string.seek(0)
                temp_table = temp_string.read()
                temp_string.close()
            else:
                fn.seek(self.sfs.chunksize * self._pointer_to_pointer_table + 0x138)
                temp_table = fn.read(self.sfs.usable_chunk)
            self.pointers = (
                np.frombuffer(
                    temp_table[: self.size_in_chunks * 4], dtype="uint32"
                ).astype(np.int64)
                * self.sfs.chunksize
                + 0x138
            )

    def read_piece(self, offset, length):
        """Read and returns raw byte string of the file without applying
        any decompression.

        Arguments:
        offset: seek value
        length: length of the data counting from the offset

        Returns:
        io.ByteIO object
        """
        data = io.BytesIO()
        # first block index:
        fb_idx = offset // self.sfs.usable_chunk
        # first block offset:
        fbo = offset % self.sfs.usable_chunk
        # last block index:
        lb_idx = (offset + length) // self.sfs.usable_chunk
        # last block cut off:
        lbco = (offset + length) % self.sfs.usable_chunk
        with open(self.sfs.filename, "rb") as fn:
            if fb_idx != lb_idx:
                fn.seek(self.pointers[fb_idx] + fbo)
                data.write(fn.read(self.sfs.usable_chunk - fbo))
                for i in self.pointers[fb_idx + 1 : lb_idx]:
                    fn.seek(i)
                    data.write(fn.read(self.sfs.usable_chunk))
                if lbco > 0:
                    fn.seek(self.pointers[lb_idx])
                    data.write(fn.read(lbco))
            else:
                fn.seek(self.pointers[fb_idx] + fbo)
                data.write(fn.read(length))
        data.seek(0)
        return data.read()

    def _iter_read_chunks(self, first=0):
        """Generate and return iterator for reading and returning
        sfs internal file in chunks.

        By default it creates iterator for whole file, however
        with kwargs 'first' and 'chunks' the range of chunks
        for iterator can be set.

        Keyword arguments:
        first -- the index of first chunk from which to read. (default 0)
        chunks -- the number of chunks to read. (default False)
        """
        last = self.size_in_chunks
        with open(self.sfs.filename, "rb") as fn:
            for idx in range(first, last - 1):
                fn.seek(self.pointers[idx])
                yield fn.read(self.sfs.usable_chunk)
            fn.seek(self.pointers[last - 1])
            last_stuff = self.size % self.sfs.usable_chunk
            if last_stuff != 0:
                yield fn.read(last_stuff)
            else:
                yield fn.read(self.sfs.usable_chunk)

    def setup_compression_metadata(self):
        """parse and setup the number of compression chunks

        and uncompressed chunk size as class attributes.

        Sets up attributes:
        self.uncompressed_blk_size, self.no_of_compr_blk

        """
        with open(self.sfs.filename, "rb") as fn:
            fn.seek(self.pointers[0])
            # AACS signature, uncompressed size, undef var, number of blocks
            aacs, uc_size, _, n_of_blocks = strct_unp("<IIII", fn.read(16))
        if aacs == 0x53434141:  # AACS as string
            self.uncompressed_blk_size = uc_size
            self.no_of_compr_blk = n_of_blocks
        else:
            raise ValueError(
                """The file is marked to be compressed,
but compression signature is missing in the header. Aborting...."""
            )

    def _iter_read_compr_chunks(self):
        """Generate and return reader and decompressor iterator
        for compressed with zlib compression sfs internal file.

        Returns:
        iterator of decompressed data chunks.
        """

        offset = 0x80  # the 1st compression block header
        for dummy1 in range(self.no_of_compr_blk):
            cpr_size = strct_unp("<I12x", self.read_piece(offset, 16))[0]
            # cpr_size, dum_size, dum_unkn, dum_size2 = strct_unp('<IIII',...
            # dum_unkn is probably some kind of checksum but
            # none of known (crc16, crc32, adler32) algorithm could match.
            # dum_size2 == cpr_size + 0x10 which have no use...
            # dum_size, which is decompressed size, also have no use...
            # as it is the same in file compression_header
            offset += 16
            raw_string = self.read_piece(offset, cpr_size)
            offset += cpr_size
            yield unzip_block(raw_string)

    def get_iter_and_properties(self):
        """Generate and return the iterator of data chunks and
        properties of such chunks such as size and count.

        Method detects if data is compressed and uses iterator with
        decompression involved, else uses simple iterator of chunks.

        Returns:
            (iterator, chunk_size, number_of_chunks)
        """
        if self.sfs.compression == "None":
            return self._iter_read_chunks(), self.sfs.usable_chunk, self.size_in_chunks
        elif self.sfs.compression == "zlib":
            return (
                self._iter_read_compr_chunks(),
                self.uncompressed_blk_size,
                self.no_of_compr_blk,
            )
        else:
            raise RuntimeError(
                "file",
                str(self.sfs.filename),
                " is compressed by not known and not",
                "implemented algorithm.\n Aborting...",
            )

    def get_as_BytesIO_string(self):
        """Get the whole file as io.BytesIO object (in memory!)."""
        data = io.BytesIO()
        data.write(b"".join(self.get_iter_and_properties()[0]))
        return data


class SFS_reader:
    """Class to read sfs file.
    SFS is AidAim software's(tm) single file system.
    The class provides basic reading capabilities of such container.
    It is capable to read compressed data in zlib, but
    SFS can contain other compression which is not implemented here.
    It is also not able to read encrypted sfs containers.

    This class can be used stand alone or inherited in construction of
    file readers using sfs technology.

    Attributes
    ----------
    filename

    """

    def __init__(self, filename):
        self.filename = filename
        # read the file header
        with open(filename, "rb") as fn:
            a = fn.read(8)
            if a != b"AAMVHFSS":
                raise TypeError("file '{0}' is not SFS container".format(filename))
            fn.seek(0x124)  # this looks to be version, as float value is always
            # nicely rounded and at older bcf versions (<1.9) it was 2.40,
            # at new (v2) - 2.60
            version, self.chunksize = strct_unp("<fI", fn.read(8))
            self.sfs_version = "{0:4.2f}".format(version)
            self.usable_chunk = self.chunksize - 32
            fn.seek(0x140)
            # the sfs tree and number of the items / files + directories in it,
            # and the number in chunks of whole sfs:
            self.tree_address, self.n_tree_items, self.sfs_n_of_chunks = strct_unp(
                "<III", fn.read(12)
            )
        self._setup_vfs()

    def _setup_vfs(self):
        """Setup the virtual file system tree represented as python dictionary
        with values populated with SFSTreeItem instances

        See also
        --------
        SFSTreeItem
        """
        with open(self.filename, "rb") as fn:
            # check if file tree do not exceed one chunk:
            n_file_tree_chunks = ceil(
                (self.n_tree_items * 0x200) / (self.chunksize - 0x20)
            )
            if n_file_tree_chunks == 1:
                # file tree do not exceed one chunk in bcf:
                fn.seek(self.chunksize * self.tree_address + 0x138)
                raw_tree = fn.read(0x200 * self.n_tree_items)
            else:
                temp_str = io.BytesIO()
                tree_address = self.tree_address
                tree_items_in_chunk = (self.chunksize - 0x20) // 0x200
                for i in range(n_file_tree_chunks):
                    # jump to tree/list address:
                    fn.seek(self.chunksize * tree_address + 0x118)
                    # next tree/list address:
                    tree_address = strct_unp("<I", fn.read(4))[0]
                    fn.seek(28, 1)
                    temp_str.write(fn.read(tree_items_in_chunk * 0x200))
                temp_str.seek(0)
                raw_tree = temp_str.read(self.n_tree_items * 0x200)
                temp_str.close()
            temp_item_list = [
                SFSTreeItem(raw_tree[i * 0x200 : (i + 1) * 0x200], self)
                for i in range(self.n_tree_items)
            ]
            # temp list with parents of items
            paths = [[h.parent] for h in temp_item_list]
        # checking the compression header which can be different per file:
        self._check_the_compresion(temp_item_list)
        if self.compression == "zlib":
            for c in temp_item_list:
                if not c.is_dir:
                    c.setup_compression_metadata()
        # convert the items to virtual file system tree
        dict_tree = self._flat_items_to_dict(paths, temp_item_list)
        # and finally set the Virtual file system:
        self.vfs = dict_tree["root"]

    def _flat_items_to_dict(self, paths, temp_item_list):
        """place items from flat list into dictionary tree
        of virtual file system
        """
        while not all(g[-1] == -1 for g in paths):
            for f in range(len(paths)):
                if paths[f][-1] != -1:
                    paths[f].extend(paths[paths[f][-1]])
        names = [j.name for j in temp_item_list]
        names.append("root")  # temp root item in dictionary
        for p in paths:
            for r in range(len(p)):
                p[r] = names[p[r]]
        for p in paths:
            p.reverse()
        root = {}
        for i in range(len(temp_item_list)):
            dir_pointer = root
            for j in paths[i]:
                if j in dir_pointer:
                    dir_pointer = dir_pointer[j]
                else:
                    dir_pointer[j] = {}
                    dir_pointer = dir_pointer[j]
            if temp_item_list[i].is_dir:
                dir_pointer[temp_item_list[i].name] = {}
            else:
                dir_pointer[temp_item_list[i].name] = temp_item_list[i]
        # return dict tree:
        return root

    def _check_the_compresion(self, temp_item_list):
        """parse, check and setup the self.compression"""
        with open(self.filename, "rb") as fn:
            # Find if there is compression:
            for c in temp_item_list:
                if not c.is_dir:
                    fn.seek(c.pointers[0])
                    if fn.read(4) == b"\x41\x41\x43\x53":  # string AACS
                        self.compression = "zlib"
                    else:
                        self.compression = "None"
                    # compression is global, can't be different per file in sfs
                    break

    def get_file(self, path):
        """Return the SFSTreeItem (aka internal file) object from
        sfs container.

        Parameters
        ----------
        path : str
            Internal file path in sfs file tree. Path accepts only
            standard - forward slash for directories.

        Returns
        -------
        object : SFSTreeItem
            SFSTreeItem, which can be read into byte stream, in chunks or
            whole using objects methods.

        Examples
        --------
        to get "file" object 'kitten.png' in folder 'catz' which
        resides in root directory of sfs, you would use:

        >>> instance_of_SFSReader.get_file('catz/kitten.png')

        See Also
        --------
        SFSTreeItem
        """
        item = self.vfs

        for i in path.split("/"):
            item = item[i]
        return item


class EDXSpectrum:
    def __init__(self, spectrum):
        """
        Wrap the objectified bruker EDS spectrum xml part
        to the python object, leaving all the xml and bruker clutter behind.

        Parameters
        ----------
        spectrum : etree xml object
            etree xml object, where spectrum.attrib['Type'] should
            be 'TRTSpectrum'.
        """
        TRTHeader = spectrum.find("./TRTHeaderedClass")
        hardware_header = TRTHeader.find(
            "./ClassInstance[@Type='TRTSpectrumHardwareHeader']"
        )
        detector_header = TRTHeader.find("./ClassInstance[@Type='TRTDetectorHeader']")
        esma_header = TRTHeader.find("./ClassInstance[@Type='TRTESMAHeader']")
        # what TRT means?
        # ESMA could stand for Electron Scanning Microscope Analysis
        spectrum_header = spectrum.find("./ClassInstance[@Type='TRTSpectrumHeader']")
        xrf_header = TRTHeader.find("./ClassInstance[@Type='TRTXrfHeader']")

        # map stuff from hardware xml branch:
        self.hardware_metadata = x2d.dictionarize(hardware_header)
        self.amplification = self.hardware_metadata["Amplification"]  # USED

        # map stuff from detector xml branch
        self.detector_metadata = x2d.dictionarize(detector_header)
        self.detector_type = self.detector_metadata["Type"]  # USED

        # decode silly hidden detector layer info:
        det_l_str = self.detector_metadata["DetLayers"]
        dec_det_l_str = codecs.decode(det_l_str.encode("ascii"), "base64")
        mini_xml = ET.fromstring(unzip_block(dec_det_l_str))
        self.detector_metadata["DetLayers"] = {}  # Overwrite with dict
        for i in list(mini_xml):
            self.detector_metadata["DetLayers"][i.tag] = dict(i.attrib)

        # map stuff from esma xml branch:
        if esma_header is not None:
            self.esma_metadata = x2d.dictionarize(esma_header)
        if xrf_header is not None:
            xrf_header_dict = x2d.dictionarize(xrf_header)
            self.esma_metadata = {
                "PrimaryEnergy": xrf_header_dict["Voltage"],
                "ElevationAngle": xrf_header_dict["ExcitationAngle"],
            }
        # USED:
        self.hv = self.esma_metadata.get("PrimaryEnergy", None)
        if self.hv is None:  # pragma: no cover
            # User has reported missing PrimaryEnergy in metadata in Quantax70
            # however, this wasn't possible to reproduce in smaller test files
            # could it be an acquisition setup issue?
            _logger.warning(
                "The beam energy could not be found in the file's metadata."
            )
        self.elev_angle = self.esma_metadata["ElevationAngle"]
        date_time = gen_iso_date_time(spectrum_header)
        if date_time is not None:
            self.date, self.time = date_time

        self.spectrum_metadata = x2d.dictionarize(spectrum_header)
        self.offset = self.spectrum_metadata["CalibAbs"]
        self.scale = self.spectrum_metadata["CalibLin"]

        # main data:
        self.data = np.fromstring(spectrum.find("./Channels").text, dtype="Q", sep=",")

    def last_non_zero_channel(self):
        """return index of last nonzero channel"""
        return self.data.nonzero()[0][-1]

    def energy_to_channel(self, energy, kV=True):
        """convert energy to channel index,
        optional kwarg 'kV' (default: True) should be set to False
        if given energy units is in V"""
        if not kV:
            en_temp = energy / 1000.0
        else:
            en_temp = energy
        return int(round((en_temp - self.offset) / self.scale))


class HyperHeader:
    """Wrap Bruker HyperMaping xml header into python object.

    Arguments:
    xml_str -- the uncompressed to be provided with extracted Header xml
    from bcf.
    indexes -- list of indexes of available datasets

    Methods:
    estimate_map_channels, estimate_map_depth

    If Bcf is version 2, the bcf can contain stacks
    of hypermaps - thus header part  can contain multiply sum eds spectras
    and it's metadata per hypermap slice which can be selected using index.
    Bcf can record number of images from different single dimensional value
    detectors (BSE, SEI, ARGUS, etc...). images representing signals are
    internally ordered and right signal image can be accessed using image
    index (do not confuse with dataset index).
    """

    def __init__(self, xml_str, indexes, instrument=None):
        root = ET.fromstring(xml_str)
        root = root.find("./ClassInstance[@Type='TRTSpectrumDatabase']")
        try:
            self.name = str(root.attrib["Name"])
        except KeyError:
            self.name = "Undefined"
            _logger.info("hypermap have no name. Giving it 'Undefined' name")
        hd = root.find("./Header")
        self.date, self.time = gen_iso_date_time(hd)
        self.version = int(literal_eval(hd.find("./FileVersion").text))
        # fill the sem and stage attributes:
        self._set_microscope(root)
        self._set_mode(instrument)
        self._set_images(root)
        self.elements = {}
        self._set_elements(root)
        self.line_counter = literal_eval(root.find("./LineCounter").text)
        self.channel_count = int(root.find("./ChCount").text)
        self.mapping_count = int(root.find("./DetectorCount").text)
        # self.channel_factors = {}
        self.spectra_data = {}
        self._set_sum_edx(root, indexes)

    def _set_microscope(self, root):
        """set microscope metadata from objectified xml part (TRTSEMData,
        TRTSEMStageData, TRTDSPConfiguration).

        BCF can contain basic parameters of SEM column, and optionally
        the stage. This metadata can be not fully or at all available to
        Esprit and saved into bcf file as it depends from license and
        the link and implementation state between the microscope's
        software and Bruker system.
        """

        semData = root.find("./ClassInstance[@Type='TRTSEMData']")
        self.sem_metadata = x2d.dictionarize(semData)
        # parse values for use in hspy metadata:
        self.hv = self.sem_metadata.get("HV", None)  # in kV
        # image/hypermap resolution in um/pixel:
        if "DX" in self.sem_metadata:
            self.units = "µm"
        else:
            self.units = "pix"
        self.x_res = self.sem_metadata.get("DX", 1.0)
        self.y_res = self.sem_metadata.get("DY", 1.0)
        # stage position:
        semStageData = root.find("./ClassInstance[@Type='TRTSEMStageData']")
        self.stage_metadata = x2d.dictionarize(semStageData)
        # DSP configuration (always present, part of Bruker system):
        DSPConf = root.find("./ClassInstance[@Type='TRTDSPConfiguration']")
        self.dsp_metadata = x2d.dictionarize(DSPConf)

    def _set_mode(self, instrument=None):
        if instrument is not None:
            self.mode = instrument
        else:
            self.mode = guess_mode(self.hv)

    def get_acq_instrument_dict(self, detector=False, **kwargs):
        """return python dictionary with acquisition instrument
        mandatory data
        """
        acq_inst = {}
        if self.hv is not None:
            acq_inst["beam_energy"] = self.hv
        if "Mag" in self.sem_metadata:
            acq_inst["magnification"] = self.sem_metadata["Mag"]
        if detector:
            eds_metadata = self.get_spectra_metadata(**kwargs)
            det = gen_detector_node(eds_metadata)
            det["EDS"]["real_time"] = self.calc_real_time()
            acq_inst["Detector"] = det
            # In case of XRF, the primary energy is only defined in
            # the spectrum metadata
            if eds_metadata.hv is not None:
                acq_inst["beam_energy"] = eds_metadata.hv

        return acq_inst

    def _parse_image(self, xml_node, overview=False):
        """parse image from bruker xml image node."""
        if overview:
            overlay_node = xml_node.find(
                "./ChildClassInstances"
                "/ClassInstance["
                # "@Type='TRTRectangleOverlayElement' and "
                "@Name='Map']/TRTSolidOverlayElement/"
                "TRTBasicLineOverlayElement/TRTOverlayElement"
            )
            if overlay_node is not None:
                # Currently not tested
                # waiting on small test files to be provided by users
                # See https://github.com/hyperspy/rosettasciio/pull/383
                overlay_dict = x2d.dictionarize(overlay_node)
                over_rect = overlay_dict["TRTOverlayElement"]["Rect"]
                x1, x2 = over_rect["Left"] * self.x_res, over_rect["Right"] * self.x_res
                y1, y2 = over_rect["Top"] * self.y_res, over_rect["Bottom"] * self.y_res
                width, height = x2 - x1, y2 - y1
                md_over_dict = {
                    "class": "Rectangles",
                    "offsets": ([x1 + width / 2, y1 + height / 2],),
                    "widths": (width,),
                    "heights": (height,),
                    "color": "yellow",
                    "linewidth": 2,
                    "facecolor": "none",
                }
        image = Container()
        image.width = int(xml_node.find("./Width").text)  # in pixels
        image.height = int(xml_node.find("./Height").text)  # in pixels
        # in bytes ('u1','u2','u4')
        image.dtype = "u" + xml_node.find("./ItemSize").text
        image.plane_count = int(xml_node.find("./PlaneCount").text)
        image.images = []
        for i in range(image.plane_count):
            img = xml_node.find("./Plane" + str(i))
            raw = codecs.decode((img.find("./Data").text).encode("ascii"), "base64")
            array1 = np.frombuffer(raw, dtype=image.dtype)
            if any(array1):
                item = self.gen_hspy_item_dict_basic()
                data = array1.reshape((image.height, image.width))
                desc = img.find("./Description")
                item["data"] = data
                item["axes"][0]["size"] = image.height
                item["axes"][1]["size"] = image.width
                item["metadata"]["Signal"] = {}
                item["metadata"]["General"] = {}
                if desc is not None:
                    item["metadata"]["General"]["title"] = str(desc.text)
                if overview and (overlay_node is not None):
                    item["metadata"]["Markers"] = {"overview": md_over_dict}
                image.images.append(item)
        return image

    def _set_images(self, root):
        """Wrap objectified xml part with image to class attributes
        for self.image.
        """
        image_nodes = root.findall("./ClassInstance[@Type='TRTImageData']")
        for n in image_nodes:
            if not (n.get("Name")):
                image_node = n
        self.image = self._parse_image(image_node)
        if self.version == 2:
            overview_node = root.findall(
                "./ClassInstance[@Type='TRTContainerClass']"
                "/ChildClassInstances"
                "/ClassInstance["
                # "@Type='TRTContainerClass' and "
                "@Name='OverviewImages']"
                "/ChildClassInstances"
                "/ClassInstance[@Type='TRTImageData']"
            )
            if len(overview_node) > 0:  # in case there is no image
                self.overview = self._parse_image(overview_node[0], overview=True)

    def _set_elements(self, root):
        """wrap objectified xml part with selection of elements to
        self.elements list
        """
        try:
            elements = root.find(
                "./ClassInstance[@Type='TRTContainerClass']"
                "/ChildClassInstances"
                "/ClassInstance[@Type='TRTElementInformationList']"
                "/ClassInstance[@Type='TRTSpectrumRegionList']"
                "/ChildClassInstances"
            )
            for j in elements.findall("./ClassInstance[@Type='TRTSpectrumRegion']"):
                tmp_d = x2d.dictionarize(j)
                # In case no information on the specific selected X-ray line is
                # available, assume it is a 'Ka' line, reflecting the fact that element
                # tables in Bruker Esprit (1.9 and 2.1) store a single K line for Li-Al
                self.elements[tmp_d["XmlClassName"]] = {
                    "line": tmp_d.get("Line", "Ka"),
                    "energy": tmp_d.get("Energy"),
                }
        except AttributeError:
            _logger.info("no element selection present in the spectra..")

    def _set_sum_edx(self, root, indexes):
        for i in indexes:
            spec_node = root.find("./SpectrumData{0}/ClassInstance".format(str(i)))
            self.spectra_data[i] = EDXSpectrum(spec_node)

    def get_consistent_min_channels(self, index=0):
        """Estimate consistent minimal size of energy axis by comparing energy
        at last recorded channel vs electron beam potential and return channel
        number corresponding to least energy. This method is safe to use with
        sliced datasets (consistent between slices) which were acquired using
        the same electron potential.

        Parameters
        ----------
        index : int
            Index of the map if multiple hypermaps are present in the same bcf.

        Returns
        -------
        optimal channel number
        """
        eds_max_energy = self.spectra_data[index].amplification / 1000  # in kV
        if self.hv and self.hv > 0 and self.hv < eds_max_energy:
            return self.spectra_data[index].energy_to_channel(self.hv)
        if self.hv is None:
            logging.warn(
                "bcf header contains no node for electron beam "
                "voltage or such node is absent.\n"
                "Using full range of recorded channels."
            )
        return self.spectra_data[index].data.shape[0]

    def estimate_map_depth(self, index=0, downsample=1, for_numpy=False):
        """Estimate minimal dtype of array using cumulative spectra
        of the all pixels so that no data would be truncated.

        The method estimates the value from sum eds spectra, dividing
        the maximum  energy pulse value from raster x and y and to be on the
        safe side multiplying by 2.

        Parameters
        ----------
        index : int
            Index of the hypermap if multiply hypermaps are present in the
            same bcf. (default 0)
        downsample : int
            Downsample factor. (default 1)
        for_numpy : bool
            If False produce unsigned, otherwise signed types: if hypermap
            will be loaded using the pure python function where numpy's inplace
            integer addition will be used, the dtype should be signed;
            If cython implementation will be used (default), then any returned
            dtypes can be safely unsigned. (default False)

        Returns
        -------
        depth : numpy.dtype
            numpy dtype large enough to use in final hypermap numpy array.

        """
        sum_eds = self.spectra_data[index].data
        # the most intensive peak is Bruker reference peak at 0kV:
        roof = (
            np.max(sum_eds)
            // self.image.width
            // self.image.height
            * 2
            * downsample
            * downsample
        )
        # this complicated nonsense below is due to numpy regression in adding
        # integer inplace to unsigned integer array. (python integers is
        # signed)
        if roof > 0xFF:
            if roof > 0xFFFF:
                if for_numpy and (downsample > 1):
                    if roof > 0xEFFFFFFF:
                        depth = np.int64
                    else:
                        depth = np.int32
                else:
                    depth = np.uint32
            else:
                if for_numpy and (downsample > 1):
                    if roof > 0xEFFF:
                        depth = np.int32
                    else:
                        depth = np.int16
                else:
                    depth = np.uint16
        else:
            if for_numpy and (downsample > 1):
                if roof > 0xEF:
                    depth = np.int16
                else:
                    depth = np.int8
            else:
                depth = np.uint8
        return depth

    def get_spectra_metadata(self, index=0):
        """return objectified xml with spectra metadata
        Arguments:
        index -- index of hypermap/spectra (default 0)
        """
        return self.spectra_data[index]

    def calc_real_time(self):
        """calculate and return real time for whole hypermap
        in seconds
        """
        line_cnt_sum = np.sum(self.line_counter, dtype=float)
        line_avg = self.dsp_metadata["LineAverage"]
        pix_avg = self.dsp_metadata["PixelAverage"]
        pix_time = self.dsp_metadata["PixelTime"]
        width = self.image.width
        real_time = line_cnt_sum * line_avg * pix_avg * pix_time * width * 1e-6
        return float(real_time)

    def gen_hspy_item_dict_basic(self):
        i = {
            "axes": [
                {
                    "name": "height",
                    "offset": 0,
                    "scale": self.y_res,
                    "units": self.units,
                },
                {
                    "name": "width",
                    "offset": 0,
                    "scale": self.x_res,
                    "units": self.units,
                },
            ],
            "metadata": {
                "Acquisition_instrument": {self.mode: self.get_acq_instrument_dict()},
                "Sample": {"name": self.name},
            },
            "original_metadata": {
                "Microscope": self.sem_metadata,
                "DSP Configuration": self.dsp_metadata,
                "Stage": self.stage_metadata,
            },
            "mapping": get_mapping(self.mode),
        }
        return i


class BCF_reader(SFS_reader):
    """Class to read bcf (Bruker hypermapping) file.

    Inherits SFS_reader and all its attributes and methods.

    Attributes:
    filename

    Methods:
    check_index_valid, parse_hypermap

    The class instantiates HyperHeader class as self.header attribute
    where all metadata, sum eds spectras, (SEM) images are stored.
    """

    def __init__(self, filename, instrument=None):
        SFS_reader.__init__(self, filename)
        header_file = self.get_file("EDSDatabase/HeaderData")
        self.available_indexes = []
        # get list of presented indexes from file tree of binary sfs container
        # while looking for file names containing the hypercube data:
        for i in self.vfs["EDSDatabase"].keys():
            if "SpectrumData" in i:
                self.available_indexes.append(int(i[-1]))
        self.def_index = min(self.available_indexes)
        header_bytes = header_file.get_as_BytesIO_string().getvalue()
        sanitized_bytes = sanitize_msxml_float(header_bytes)
        self.header = HyperHeader(
            sanitized_bytes, self.available_indexes, instrument=instrument
        )
        self.hypermap = {}

    def check_index_valid(self, index):
        """check and return if index is valid"""
        if not isinstance(index, int):
            raise TypeError("provided index should be integer")
        if index not in self.available_indexes:
            raise IndexError(
                "requisted index is not in the list of available indexes. "
                "Available maps are under indexes: {0}".format(
                    str(self.available_indexes)
                )
            )
        return index

    def parse_hypermap(self, index=None, downsample=1, cutoff_at_kV=None, lazy=False):
        """Unpack the Delphi/Bruker binary spectral map and return
        numpy array in memory efficient way.

        Pure python/numpy implementation -- slow, or
        cython/memoryview/numpy implementation if compiled and present
        (fast) is used.

        Parameters
        ----------
        index : None or int
            The index of hypermap in bcf if there is more than one
            hyper map in file.
        downsample : int
            Downsampling factor. Differently than block_reduce from
            skimage.measure, the parser populates reduced array by suming
            results of pixels, thus having lower memory requiriments. Default
            is 1.
        cutoff_at_kV : None, float, int or str
            Value or method to truncate the array at energy in kV. Helps reducing size of the returned
            array. Default value is None (does not truncate). Numerical value should be in kV.
            Two methods for automatic cutoff is available:
              "zealous" - truncates to the last non zero channel (should not be used for stacks).
              "auto" - truncates to hv of electron microscope (good for stacks if hv is consistent).
        lazy : bool
            It True, returns dask.array otherwise a numpy.array. Default is
            False.

        Returns
        -------
        result : numpy.ndarray or dask.array.array
            Bruker hypermap, with (y,x,E) shape.
        """
        if index is None:
            index = self.def_index

        if type(cutoff_at_kV) in (int, float):
            eds = self.header.spectra_data[index]
            n_channels = eds.energy_to_channel(cutoff_at_kV)
        elif cutoff_at_kV == "zealous":
            n_channels = self.header.spectra_data[index].last_non_zero_channel() + 1
        elif cutoff_at_kV == "auto":
            n_channels = self.header.get_consistent_min_channels(index=index)
        else:  # None
            n_channels = self.header.spectra_data[index].data.size

        shape = (
            ceil(self.header.image.height / downsample),
            ceil(self.header.image.width / downsample),
            n_channels,
        )
        sfs_file = SFS_reader(self.filename)
        vrt_file_hand = sfs_file.get_file("EDSDatabase/SpectrumData" + str(index))
        if fast_unbcf:
            parse_func = unbcf_fast.parse_to_numpy
            dtype = self.header.estimate_map_depth(
                index=index, downsample=downsample, for_numpy=False
            )
        else:
            parse_func = py_parse_hypermap
            dtype = self.header.estimate_map_depth(
                index=index, downsample=downsample, for_numpy=True
            )
        if lazy:
            import dask.array as da

            value = dask.delayed(parse_func)(
                vrt_file_hand, shape, dtype, downsample=downsample
            )
            result = da.from_delayed(value, shape=shape, dtype=dtype)
        else:
            result = parse_func(vrt_file_hand, shape, dtype, downsample=downsample)
        return result

    def add_filename_to_general(self, item):
        """hypy helper method"""
        item["metadata"]["General"]["original_filename"] = basename(self.filename)


def spx_reader(filename, lazy=False):
    with open(filename, "br") as fn:
        xml_str = fn.read()
    root = ET.fromstring(xml_str)
    sp_node = root.find("./ClassInstance[@Type='TRTSpectrum']")
    try:
        name = str(sp_node.attrib["Name"])
    except KeyError:
        name = "Undefinded"
        _logger.info("spectra have no name. Giving it 'Undefined' name")
    spectrum = EDXSpectrum(sp_node)
    mode = guess_mode(spectrum.hv)
    results_xml = sp_node.find("./ClassInstance[@Type='TRTResult']")
    elements_xml = sp_node.find("./ClassInstance[@Type='TRTPSEElementList']")
    hy_spec = {
        "data": spectrum.data,
        "axes": [
            {
                "name": "Energy",
                "size": len(spectrum.data),
                "offset": spectrum.offset,
                "scale": spectrum.scale,
                "units": "keV",
                "navigate": False,
            }
        ],
        "metadata":
        # where is no way to determine what kind of instrument was used:
        # TEM or SEM
        {
            "Acquisition_instrument": {
                mode: {
                    "Detector": gen_detector_node(spectrum),
                    "beam_energy": spectrum.hv,
                }
            },
            "General": {
                "original_filename": basename(filename),
                "title": "EDX",
                "date": spectrum.date,
                "time": spectrum.time,
            },
            "Sample": {"name": name},
            "Signal": {
                "signal_type": "EDS_%s" % mode,
                "quantity": "X-rays (Counts)",
            },
        },
        "original_metadata": {
            "Hardware": spectrum.hardware_metadata,
            "Detector": spectrum.detector_metadata,
            "Analysis": spectrum.esma_metadata,
            "Spectrum": spectrum.spectrum_metadata,
        },
    }
    if results_xml is not None:
        hy_spec["original_metadata"]["Results"] = x2d.dictionarize(results_xml)
    if elements_xml is not None:
        elem = x2d.dictionarize(elements_xml)["ChildClassInstances"]
        hy_spec["original_metadata"]["Selected_elements"] = elem
        hy_spec["metadata"]["Sample"]["elements"] = elem["XmlClassName"]
    return [hy_spec]


# dict of nibbles to struct notation for reading:
st = {1: "B", 2: "B", 4: "H", 8: "I", 16: "Q"}


def py_parse_hypermap(virtual_file, shape, dtype, downsample=1):
    """Unpack the Delphi/Bruker binary spectral map and return
    numpy array in memory efficient way using pure python implementation.
    (Slow!)

    The function is long and complicated due to complexity of Delphi packed
    array.
    Whole parsing is placed in one function to reduce overhead of
    python function calls. For cleaner parsing logic, please, see
    fast cython implementation at rsciio/bruker/unbcf_fast.pyx

    The method is only meant to be used if for some
    reason c (generated with cython) version of the parser is not compiled.

    Parameters
    ----------
    virtual_file -- virtual file handle returned by SFS_reader instance
        or by object inheriting it (e.g. BCF_reader instance)
    shape -- numpy shape
    dtype -- numpy dtype
    downsample -- downsample factor

    note!: downsample, shape and dtype are interconnected and needs
    to be properly calculated otherwise wrong output or segfault
    is expected

    Returns
    -------
    numpy array of bruker hypermap, with (y, x, E) shape.
    """
    iter_data, size_chnk = virtual_file.get_iter_and_properties()[:2]
    dwn_factor = downsample
    max_chan = shape[2]
    buffer1 = next(iter_data)
    height, width = strct_unp("<ii", buffer1[:8])
    # hyper map as very flat array:
    vfa = np.zeros(shape[0] * shape[1] * shape[2], dtype=dtype)
    offset = 0x1A0
    size = size_chnk
    for line_cnt in range(height):
        if (offset + 4) >= size:
            size = size_chnk + size - offset
            buffer1 = buffer1[offset:] + next(iter_data)
            offset = 0
        line_head = strct_unp("<i", buffer1[offset : offset + 4])[0]
        offset += 4
        for dummy1 in range(line_head):
            if (offset + 22) >= size:
                size = size_chnk + size - offset
                buffer1 = buffer1[offset:] + next(iter_data)
                offset = 0
            # the pixel header contains such information:
            # x index of pixel (uint32);
            # number of channels for whole mapping (unit16);
            # number of channels for pixel (uint16);
            # dummy placehollder (same value in every known bcf) (32bit);
            # flag distinguishing packing data type (16bit):
            #    0 - 16bit packed pulses, 1 - 12bit packed pulses,
            #    >1 - instructively packed spectra;
            # value which sometimes shows the size of packed data (uint16);
            # number of pulses if pulse data are present (uint16) or
            #      additional pulses to the instructively packed data;
            # packed data size (32bit) (without additional pulses) \
            #       next header is after that amount of bytes;
            (
                x_pix,
                chan1,
                chan2,
                dummy1,
                flag,
                dummy_size1,
                n_of_pulses,
                data_size2,
            ) = strct_unp("<IHHIHHHI", buffer1[offset : offset + 22])
            pix_idx = (x_pix // dwn_factor) + (
                ceil(width / dwn_factor) * (line_cnt // dwn_factor)
            )
            offset += 22
            if (offset + data_size2) >= size:
                buffer1 = buffer1[offset:] + next(iter_data)
                size = size_chnk + size - offset
                offset = 0
            if flag == 0:
                data1 = buffer1[offset : offset + data_size2]
                arr16 = np.frombuffer(data1, dtype=np.uint16)
                pixel = np.bincount(arr16, minlength=chan1 - 1)
                offset += data_size2
            elif flag == 1:  # and (chan1 != chan2)
                # Unpack packed 12-bit data to 16-bit uints:
                data1 = buffer1[offset : offset + data_size2]
                switched_i2 = np.frombuffer(data1, dtype="<u2").copy().byteswap(True)
                data2 = (
                    np.frombuffer(switched_i2.tobytes(), dtype=np.uint8)
                    .copy()
                    .repeat(2)
                )
                mask = np.ones_like(data2, dtype=bool)
                mask[0::6] = mask[5::6] = False
                # Reinterpret expanded as 16-bit:
                # string representation of array after switch will have
                # always BE independently from endianness of machine
                exp16 = np.frombuffer(
                    data2[mask].tobytes(), dtype=">u2", count=n_of_pulses
                ).copy()
                exp16[0::2] >>= 4  # Shift every second short by 4
                exp16 &= np.uint16(0x0FFF)  # Mask all shorts to 12bit
                pixel = np.bincount(exp16, minlength=chan1 - 1)
                offset += data_size2
            else:  # flag > 1
                # Unpack instructively packed data to pixel channels:
                pixel = []
                the_end = offset + data_size2 - 4
                while offset < the_end:
                    # this would work on py3
                    # size_p, channels = buffer1[offset:offset + 2]
                    # this is needed on py2:
                    size_p, channels = strct_unp("<BB", buffer1[offset : offset + 2])
                    offset += 2
                    if size_p == 0:
                        pixel += channels * [0]
                    else:
                        gain = strct_unp(
                            "<" + st[size_p * 2], buffer1[offset : offset + size_p]
                        )[0]
                        offset += size_p
                        if size_p == 1:
                            # special case with nibble switching
                            length = ceil(channels / 2)
                            # valid py3 code
                            # a = list(buffer1[offset:offset + length])
                            # this have to be used on py2:
                            a = strct_unp(
                                "<" + "B" * length, buffer1[offset : offset + length]
                            )
                            g = []
                            for i in a:
                                g += (i & 0x0F) + gain, (i >> 4) + gain
                            pixel += g[:channels]
                        else:
                            length = int(channels * size_p / 2)
                            temp = strct_unp(
                                "<" + channels * st[size_p],
                                buffer1[offset : offset + length],
                            )
                            pixel += [l_ + gain for l_ in temp]
                        offset += length
                if chan2 < chan1:
                    rest = chan1 - chan2
                    pixel += rest * [0]
                # additional data size:
                if n_of_pulses > 0:
                    add_s = strct_unp("<I", buffer1[offset : offset + 4])[0]
                    offset += 4
                    if (offset + add_s) >= size:
                        buffer1 = buffer1[offset:] + next(iter_data)
                        size = size_chnk + size - offset
                        offset = 0
                    # the additional pulses:
                    add_pulses = strct_unp(
                        "<" + "H" * n_of_pulses, buffer1[offset : offset + add_s]
                    )
                    offset += add_s
                    for i in add_pulses:
                        pixel[i] += 1
                else:
                    offset += 4
            # if no downsampling is needed, or if it is first
            # pixel encountered with downsampling on, then
            # use assignment, which is ~4 times faster, than inplace add
            if max_chan < chan1:  # if pixel have more channels than we need
                chan1 = max_chan
            if dwn_factor == 1:
                vfa[max_chan * pix_idx : chan1 + max_chan * pix_idx] = pixel[:chan1]
            else:
                vfa[max_chan * pix_idx : chan1 + max_chan * pix_idx] += pixel[:chan1]
    vfa.resize((ceil(height / dwn_factor), ceil(width / dwn_factor), max_chan))
    # check if array is signed, and convert to unsigned
    if str(vfa.dtype)[0] == "i":
        new_dtype = "".join(["u", str(vfa.dtype)])
        vfa.dtype = new_dtype
    return vfa


def file_reader(
    filename,
    lazy=False,
    select_type=None,
    index=None,
    downsample=1,
    cutoff_at_kV=None,
    instrument=None,
):
    """
    Read a Bruker ``.bcf`` or ``.spx`` file.

    Parameters
    ----------
    %s
    %s
    select_type : ``'spectrum_image'``, ``'image'`` or None, default=None
        If specified, only the corresponding type of data, either spectrum
        image or image, is returned.
        By default (None), all data are loaded.
    index : int, None or str, default=None
        Allow to select the index of the dataset in the ``.bcf`` file, which
        can contain several datasets. The default value (``None``) results in
        loading the first dataset. When set to ``'all'``, all available datasets
        will be loaded and returned as separate signals.
    downsample : int, default=1
        The downsampling ratio of a hyperspectral array (height and width only),
        can be an integer >=1, where value of 1 results in no downsampling.
        The underlying method of downsampling is unchangeable: sum. Differently
        than :py:func:``skimage.measure.block_reduce``, it is memory efficient
        as it doesn't create intermediate arrays.
    cutoff_at_kV : int, float, ``'zealous'``, ``'auto'`` or None, default=None
        It can be used either to crop or enlarge the energy (or number of channels)
        range at max values. It can be used to conserve memory or enlarge
        the range if needed to mach the size of other file. The default value
        (``None``) does not influence the size. Numerical values should be in kV.
        ``'zealous'`` truncates to the last non zero channel (this option
        should not be used for stacks, as low beam current EDS can have a different
        last non zero channel per slice). ``'auto'`` truncates channels to SEM/TEM
        acceleration voltage or energy at last channel, depending which is smaller.
        In case the hv info is not there or hv is off (0 kV) then it falls back to
        the full channel range.
    instrument : str or None, default=None
        Can be either ``'TEM'`` or ``'SEM'``.

    %s

    Examples
    --------
    Example of loading reduced (downsampled, and with energy range cropped)
    "spectrum only" data from ``bcf`` (original shape: 80 keV EDS range (4096 channels),
    100x75 pixels; SEM acceleration voltage: 20kV):

    >>> file_reader("sample80kv.bcf", select_type='spectrum_image',
                    downsample=2, cutoff_at_kV=10)

    Load the same file with limiting array size to SEM acceleration voltage:

    >>> file_reader("sample80kv.bcf", cutoff_at_kV='auto')

    The loaded array energy dimension can by forced to be larger than the data
    recorded by setting the 'cutoff_at_kV' kwarg to higher value:

    >>> file_reader("sample80kv.bcf", cutoff_at_kV=60)

    Loading without setting ``cutoff_at_kV`` value would return data with all 4096
    channels. Note that setting ``downsample`` higher than 1 currently locks out using SEM
    images for navigation in the plotting.
    """
    ext = splitext(filename)[1][1:].lower()
    if ext == "bcf":
        to_return = bcf_reader(
            filename=filename,
            lazy=lazy,
            select_type=select_type,
            index=index,
            downsample=downsample,
            cutoff_at_kV=cutoff_at_kV,
            instrument=instrument,
        )
    elif ext == "spx":
        to_return = spx_reader(
            filename=filename,
            lazy=lazy,
        )
    else:
        raise ValueError(f"'{ext}' is not a supported extension for the bruker reader.")

    return to_return


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)


def bcf_reader(
    filename,
    lazy=False,
    select_type=None,
    index=None,  # noqa
    downsample=1,
    cutoff_at_kV=None,
    instrument=None,
):
    """
    Reads a bruker ``.bcf`` file and loads the data into the appropriate class,
    then wraps it into a list of dictionaries typically used by the
    :py:func:`hyperspy.api.load` function.

    Parameters
    ----------
    select_type : str or None, default=None
        One of: spectrum_image, image. If none specified, then function
        loads everything, else if specified, loads either just sem imagery,
        or just hyper spectral mapping data.
    index : int, None or str, default=None
        Index of dataset in bcf v2 can be None integer and 'all';
        None will select first available mapping if more than one.
        'all' will return all maps if more than one present;
        integer will return only selected map.
    downsample : int, default=1
        the downsample ratio of hyperspectral array (downsampling
        height and width only), can be integer from 1 to inf, where '1' means
        no downsampling will be applied.
    cutoff_at_kV : int, float or None, default=None
        if set (can be int of float >= 0) can be used either, to
        crop or enlarge energy range at max values.
    instrument : str or None, default=None
        Can be either 'TEM' or 'SEM'.
    """

    # objectified bcf file:
    obj_bcf = BCF_reader(filename, instrument=instrument)
    if select_type == "image":
        return bcf_images(obj_bcf)
    elif select_type == "spectrum_image":
        return bcf_hyperspectra(
            obj_bcf,
            index=index,
            downsample=downsample,
            cutoff_at_kV=cutoff_at_kV,
            lazy=lazy,
        )
    else:
        return bcf_images(obj_bcf) + bcf_hyperspectra(
            obj_bcf,
            index=index,
            downsample=downsample,
            cutoff_at_kV=cutoff_at_kV,
            lazy=lazy,
        )


def bcf_images(obj_bcf):
    """Returns list of dict with sem images and metadata."""
    images_list = []
    for img in obj_bcf.header.image.images:
        obj_bcf.add_filename_to_general(img)
        images_list.append(img)
    if hasattr(obj_bcf.header, "overview"):
        for img2 in obj_bcf.header.overview.images:
            obj_bcf.add_filename_to_general(img2)
            images_list.append(img2)
    return images_list


def bcf_hyperspectra(
    obj_bcf,
    index=None,
    downsample=None,
    cutoff_at_kV=None,
    lazy=False,  # noqa
):
    """Returns list of dict with eds hyperspectra and metadata."""
    global warn_once
    if (fast_unbcf is False) and warn_once:
        _logger.warning(
            """unbcf_fast library is not present...
Parsing BCF with Python-only backend, which is slow... please wait.
If parsing is uncomfortably slow, first install cython, then reinstall RosettaSciIO.
For more information, check the 'Installing RosettaSciIO' section in the documentation."""
        )
        warn_once = False
    if index is None:
        indexes = [obj_bcf.def_index]
    elif index == "all":
        indexes = obj_bcf.available_indexes
    else:
        indexes = [obj_bcf.check_index_valid(index)]
    hyperspectra = []
    mode = obj_bcf.header.mode
    mapping = get_mapping(mode)
    for index in indexes:
        hypermap = obj_bcf.parse_hypermap(
            index=index, downsample=downsample, cutoff_at_kV=cutoff_at_kV, lazy=lazy
        )
        eds_metadata = obj_bcf.header.get_spectra_metadata(index=index)
        hyperspectra.append(
            {
                "data": hypermap,
                "axes": [
                    {
                        "name": "height",
                        "size": hypermap.shape[0],
                        "offset": 0,
                        "scale": obj_bcf.header.y_res * downsample,
                        "units": obj_bcf.header.units,
                        "navigate": True,
                    },
                    {
                        "name": "width",
                        "size": hypermap.shape[1],
                        "offset": 0,
                        "scale": obj_bcf.header.y_res * downsample,
                        "units": obj_bcf.header.units,
                        "navigate": True,
                    },
                    {
                        "name": "Energy",
                        "size": hypermap.shape[2],
                        "offset": eds_metadata.offset,
                        "scale": eds_metadata.scale,
                        "units": "keV",
                        "navigate": False,
                    },
                ],
                "metadata":
                # where is no way to determine what kind of instrument was used:
                # TEM or SEM
                {
                    "Acquisition_instrument": {
                        mode: obj_bcf.header.get_acq_instrument_dict(
                            detector=True, index=index
                        )
                    },
                    "General": {
                        "original_filename": basename(obj_bcf.filename),
                        "title": "EDX",
                        "date": obj_bcf.header.date,
                        "time": obj_bcf.header.time,
                    },
                    "Sample": {
                        "name": obj_bcf.header.name,
                        "elements": sorted(list(obj_bcf.header.elements)),
                        "xray_lines": sorted(gen_elem_list(obj_bcf.header.elements)),
                    },
                    "Signal": {
                        "signal_type": "EDS_%s" % mode,
                        "quantity": "X-rays (Counts)",
                    },
                },
                "original_metadata": {
                    "Hardware": eds_metadata.hardware_metadata,
                    "Detector": eds_metadata.detector_metadata,
                    "Analysis": eds_metadata.esma_metadata,
                    "Spectrum": eds_metadata.spectrum_metadata,
                    "DSP Configuration": obj_bcf.header.dsp_metadata,
                    "Line counter": obj_bcf.header.line_counter,
                    "Stage": obj_bcf.header.stage_metadata,
                    "Microscope": obj_bcf.header.sem_metadata,
                },
                "mapping": mapping,
            }
        )
    return hyperspectra


def gen_elem_list(the_dict):
    return ["_".join([i, parse_line(the_dict[i]["line"])]) for i in the_dict]


def parse_line(line_string):
    """standardize line description.

    Bruker saves line description in all caps
    and omits the type if only one exists instead of
    using alfa"""
    if len(line_string) == 1:
        line_string = line_string + "a"
    elif len(line_string) > 2:
        line_string = line_string[:2]
    return line_string.capitalize()


def get_mapping(mode):
    return {
        "Stage.Rotation": ("Acquisition_instrument.%s.Stage.rotation" % mode, None),
        "Stage.Tilt": ("Acquisition_instrument.%s.Stage.tilt_alpha" % mode, None),
        "Stage.X": ("Acquisition_instrument.%s.Stage.x" % mode, None),
        "Stage.Y": ("Acquisition_instrument.%s.Stage.y" % mode, None),
        "Stage.Z": ("Acquisition_instrument.%s.Stage.z" % mode, None),
    }


def guess_mode(hv):
    """there is no way to determine what kind of instrument
    was used from metadata: TEM or SEM.
    However simple guess can be made using the acceleration
    voltage, assuming that SEM is <= 30kV or TEM is >30kV"""
    if hv is not None and hv > 30.0:
        mode = "TEM"
    else:
        mode = "SEM"

    if hv is not None:
        _logger.info(
            "Guessing that the acquisition instrument is %s " % mode
            + "because the beam energy is %i keV. If this is wrong, " % hv
            + "please provide the right instrument using the 'instrument' "
            + "keyword."
        )
    return mode


def gen_detector_node(spectrum):
    eds_dict = {
        "EDS": {
            "elevation_angle": spectrum.elev_angle,
            "detector_type": spectrum.detector_type,
        }
    }
    if "AzimutAngle" in spectrum.esma_metadata:
        eds_dict["EDS"]["azimuth_angle"] = spectrum.esma_metadata["AzimutAngle"]
    if "RealTime" in spectrum.hardware_metadata:
        eds_dict["EDS"]["real_time"] = spectrum.hardware_metadata["RealTime"] / 1000
        eds_dict["EDS"]["live_time"] = spectrum.hardware_metadata["LifeTime"] / 1000
    return eds_dict


def gen_iso_date_time(node):
    date_xml = node.find("./Date")
    time_xml = node.find("./Time")
    if date_xml is not None:
        dt = datetime.strptime(
            " ".join([date_xml.text, time_xml.text]), "%d.%m.%Y %H:%M:%S"
        )
        date = dt.date().isoformat()
        time = dt.time().isoformat()
        return date, time
    else:
        return None



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import dask.array as da
import numpy as np

import rsciio.utils.rgb as rt


class TestRGBTools:
    def setup_method(self, method):
        self.data_c = np.ones((2, 2, 3), dtype=np.uint8, order="C")
        self.data_f = np.ones((2, 2, 3), dtype=np.uint8, order="F")
        mask = [[[0, 1, 1], [1, 0, 1]], [[1, 1, 0], [0, 0, 1]]]
        self.data_masked = np.ma.masked_array(self.data_c, mask, hard_mask=True)

    def test_rgbx2regular_array_corder_from_c(self):
        d = rt.rgbx2regular_array(self.data_c)
        assert d.flags["C_CONTIGUOUS"]

    def test_rgbx2regular_array_corder_from_c_dask(self):
        data = da.from_array(self.data_c)
        d = rt.rgbx2regular_array(data)
        assert d.flags["C_CONTIGUOUS"]

    def test_rgbx2regular_array_corder_from_f(self):
        d = rt.rgbx2regular_array(self.data_f)
        assert d.flags["C_CONTIGUOUS"]

    def test_rgbx2regular_array_corder_from_c_slices(self):
        d = rt.rgbx2regular_array(self.data_c[0:1, ...])
        assert d.flags["C_CONTIGUOUS"]
        d = rt.rgbx2regular_array(self.data_c[:, 0:1, :])
        assert d.flags["C_CONTIGUOUS"]

    def test_rgbx2regular_array_cordermask_from_cmasked(self):
        d = rt.rgbx2regular_array(self.data_masked)
        assert isinstance(d, np.ma.MaskedArray)
        assert d.flags["C_CONTIGUOUS"]

    def test_rgbx2regular_array_cordermask_from_cmasked_slices(self):
        d = rt.rgbx2regular_array(self.data_masked[0:1, ...])
        assert d.flags["C_CONTIGUOUS"]
        assert isinstance(d, np.ma.MaskedArray)
        d = rt.rgbx2regular_array(self.data_masked[:, 0:1, :])
        assert d.flags["C_CONTIGUOUS"]
        assert isinstance(d, np.ma.MaskedArray)

    def test_regular_array2rgbx_corder_from_c(self):
        d = rt.regular_array2rgbx(self.data_c)
        assert d.flags["C_CONTIGUOUS"]

    def test_regular_array2rgbx_corder_from_f(self):
        d = rt.regular_array2rgbx(self.data_f)
        assert d.flags["C_CONTIGUOUS"]

    def test_regular_array2rgbx_corder_from_c_slices(self):
        d = rt.regular_array2rgbx(self.data_c[0:1, ...])
        assert d.flags["C_CONTIGUOUS"]
        d = rt.regular_array2rgbx(self.data_c[:, 0:1, :])
        assert d.flags["C_CONTIGUOUS"]

    def test_regular_array2rgbx_cordermask_from_cmasked(self):
        d = rt.regular_array2rgbx(self.data_masked)
        assert isinstance(d, np.ma.MaskedArray)
        assert d.flags["C_CONTIGUOUS"]

    def test_regular_array2rgbx_cordermask_from_cmasked_slices(self):
        d = rt.regular_array2rgbx(self.data_masked[0:1, ...])
        assert isinstance(d, np.ma.MaskedArray)
        assert d.flags["C_CONTIGUOUS"]
        d = rt.regular_array2rgbx(self.data_masked[:, 0:1, :])
        assert isinstance(d, np.ma.MaskedArray)
        assert d.flags["C_CONTIGUOUS"]



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import importlib.util
import logging
import xml.etree.ElementTree as ET
from collections import defaultdict
from copy import deepcopy
from pathlib import Path

import numpy as np
from numpy.polynomial.polynomial import polyfit

from rsciio._docstrings import FILENAME_DOC, LAZY_DOC, RETURNS_DOC

_logger = logging.getLogger(__name__)


def _error_handling_find_location(len_entry, name):
    if len_entry > 1:
        _logger.error(  # pragma: no cover
            f"File contains multiple positions to read {name} from."
            "                            "
            "The first location is chosen."
        )
    elif len_entry == 0:
        raise RuntimeError(
            f"Could not find location to read {name} from."
        )  # pragma: no cover


def _convert_float(input):
    """Handle None-values when converting strings to float."""
    if input is None:
        return None  # pragma: no cover
    else:
        return float(input)


def _remove_none_from_dict(dict_in):
    """Recursive removal of None-values from a dictionary."""
    for key, value in list(dict_in.items()):
        if isinstance(value, dict):
            _remove_none_from_dict(value)
        elif value is None:
            del dict_in[key]


def _etree_to_dict(t, only_top_lvl=False):
    """Recursive conversion from xml.etree.ElementTree to a dictionary."""
    d = {t.tag: {} if t.attrib else None}
    if not only_top_lvl:
        children = list(t)
        if children:
            dd = defaultdict(list)
            for dc in map(_etree_to_dict, children):
                for k, v in dc.items():
                    dd[k].append(v)
            d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
        if t.text:
            if children or t.attrib:
                ## in this case, the text is ignored
                ## if children=True -> text is just empty space in test data
                ## if t.attrib=True and children=False doesn't occur in test data
                pass
            else:
                d[t.tag] = t.text.strip()
    if t.attrib:
        d[t.tag].update((k, v) for k, v in t.attrib.items())
    return d


def _process_info_serialized(head):
    """Recursive processing designed for the InfoSerialized entry from the original_metadata."""
    result = {}
    if not isinstance(head, list):
        head = [head]
    save_name_for_rename = set()
    for idx, entry in enumerate(head):
        entry_name = entry["Name"]
        if entry["Groups"] is not None:
            result[entry_name] = _process_info_serialized(entry["Groups"]["Group"])
        else:
            entry_dict_old = entry["Items"]["Item"]
            entry_dict_new = dict()
            if isinstance(entry_dict_old, list):
                for item in entry_dict_old:
                    entry_dict_new[item["Name"]] = item["Value"]
            elif isinstance(entry_dict_old, dict):
                entry_dict_new[entry_dict_old["Name"]] = entry_dict_old["Value"]
            if entry_name in result:
                save_name_for_rename.add(entry_name)
                entry_name += str(idx + 1)
            result[entry_name] = entry_dict_new
    for name in save_name_for_rename:
        result[name + "1"] = result.pop(name)
    return result


class TrivistaTVFReader:
    """Class to read Trivista's .tvf-files using xml.etree.ElementTree.

    Attributes
    ----------
    data, metadata, original_metadata, axes
    """

    def __init__(
        self,
        file_path,
        use_uniform_signal_axis=False,
        glued_data_as_stack=False,
        filter_original_metadata=True,
    ):
        self._file_path = file_path
        self._use_uniform_signal_axis = use_uniform_signal_axis
        self._glued_data_as_stack = glued_data_as_stack

        (
            data_head,
            filtered_original_metadata,
            unfiltered_original_metadata,
        ) = self.parse_file_structure(filter_original_metadata)

        self._num_datasets = int(
            data_head.findall("Childs")[0].attrib["Count"]  # codespell:ignore
        )
        data, time, signal_axis = self.get_data_and_signal(data_head=data_head)

        self.axes = self.set_axes(
            axis_head=filtered_original_metadata["Document"]["InfoSerialized"],
            signal_axis=signal_axis,
            time=time,
        )

        self.data = self.reshape_data(data, self.axes)
        self.metadata = self.map_metadata(filtered_original_metadata)

        if filter_original_metadata:
            self.original_metadata = filtered_original_metadata
        else:
            self.original_metadata = unfiltered_original_metadata

    def parse_file_structure(self, filter_original_metadata):
        """Initial parse through the file to extract original metadata and the data location.

        In general the file structure looks something like this.
        - root_level
            - root_level metadata
                - FileInfoSerialized
            - Document
                - data/signal_axis
                - document metadata
                    - InfoSerialized
            - Hardware
                - hardware metadata

        Most of the usable metadata is in the InfoSerialized section.
        InfoSerialized and FileInfoSerialized need to be converted extra
        with ET.fromstring().

        The metadata from the hardware section contains information for all
        available hardware (i.e. multiple objectives even though only one is used.
        In the filtering process, the metadata of the actual used objective is extracted).

        Parameters
        ----------
        filter_original_metadata: bool
            if True, then unfiltered_original_metadata will
            contain a copy of the original_metadata before
            processing InfoSerialized and filtering the metadata
            from the Hardware section. Otherwise unfiltered_original_metadata
            will be empty.
            Independent of this parameter, the filtered and processed
            metadata is stored in filtered_original_metadata.
            This ensures that the metadata mapping doesn't
            depend on this setting.

        Returns
        -------
        data_head: ET
            file position where data/signal can be read from

        filtered_original_metadata: dict
            filtered + processed metadata

        unfiltered_original_metadata: dict
        """
        filtered_original_metadata = dict()
        unfiltered_original_metadata = dict()
        et_root = ET.parse(self._file_path).getroot()

        ## root level metadata
        filtered_original_metadata.update(_etree_to_dict(et_root, only_top_lvl=True))
        et_fileInfoSerialized = ET.fromstring(
            filtered_original_metadata["XmlMain"]["FileInfoSerialized"]
        )
        fileInfoSerialized = _etree_to_dict(et_fileInfoSerialized, only_top_lvl=False)
        filtered_original_metadata["XmlMain"]["FileInfoSerialized"] = fileInfoSerialized

        ## Documents / Document section
        data_head = et_root[1][0]
        filtered_original_metadata.update(_etree_to_dict(data_head, only_top_lvl=True))
        et_infoSerialized = ET.fromstring(
            filtered_original_metadata["Document"]["InfoSerialized"]
        )
        infoSerialized = _etree_to_dict(et_infoSerialized, only_top_lvl=False)

        ## Hardware section
        metadata_head = et_root[0]
        metadata_hardware = _etree_to_dict(metadata_head, only_top_lvl=False)

        if not filter_original_metadata:
            unfiltered_original_metadata = deepcopy(filtered_original_metadata)
            unfiltered_original_metadata["Document"]["InfoSerialized"] = deepcopy(
                infoSerialized
            )
            unfiltered_original_metadata.update(deepcopy(metadata_hardware))

        ## processing/filtering
        infoSerialized_processed = _process_info_serialized(
            infoSerialized["Info"]["Groups"]["Group"]
        )
        filtered_original_metadata["Document"]["InfoSerialized"] = (
            infoSerialized_processed
        )

        ## these methods alter metadata_hardware
        self._filter_laser_metadata(infoSerialized_processed, metadata_hardware)
        self._filter_detector_metadata(infoSerialized_processed, metadata_hardware)
        self._filter_objectives_metadata(metadata_hardware)
        self._filter_spectrometers_metadata(infoSerialized_processed, metadata_hardware)

        filtered_original_metadata.update(metadata_hardware)

        return data_head, filtered_original_metadata, unfiltered_original_metadata

    @staticmethod
    def _filter_laser_metadata(infoSerialized_processed, metadata_hardware):
        """Filter LightSources section (Laser) via wavelength if possible."""
        for laser in metadata_hardware["Hardware"]["LightSources"]["LightSource"]:
            try:
                calibration_wl = float(
                    infoSerialized_processed["Calibration"]["Laser_Wavelength"]
                )
                laser_wl = float(laser["Wavelengths"]["Value_0"])
            except KeyError:
                pass
            else:
                if np.isclose(calibration_wl, laser_wl) and not np.isclose(laser_wl, 0):
                    metadata_hardware["Hardware"]["LightSources"]["LightSource"] = laser

    @staticmethod
    def _filter_detector_metadata(infoSerialized_processed, metadata_hardware):
        """Filter Detector section via name"""
        for detector in metadata_hardware["Hardware"]["Detectors"]["Detector"]:
            if detector["Name"] == infoSerialized_processed["Detector"]["Name"]:
                metadata_hardware["Hardware"]["Detectors"]["Detector"] = detector

    @staticmethod
    def _filter_objectives_metadata(metadata_hardware):
        """Filter microscope section (objective) via isEnabled tag"""
        for microscope in metadata_hardware["Hardware"]["Microscopes"]["Microscope"]:
            for objective in microscope["Objectives"]["Objective"]:
                if objective["IsEnabled"] == "True":
                    metadata_hardware["Hardware"]["Microscopes"]["Microscope"] = (
                        microscope
                    )
                    metadata_hardware["Hardware"]["Microscopes"]["Microscope"][
                        "Objectives"
                    ]["Objective"] = objective

    @staticmethod
    def _filter_spectrometers_metadata(infoSerialized_processed, metadata_hardware):
        """Filter spectrometers via serialnumbers"""
        ## get serialnumbers for all used spectrometers
        ## contrary to the other parts, multiple spectrometers can be used
        spectrometer_serial_numbers = []
        spectrometer_serialized_list = []
        for key, val in infoSerialized_processed["Spectrometers"].items():
            spectrometer_serial_numbers.append(val["Serialnumber"])
            spectrometer_serialized_list.append(key)

        ## filter spectrometers via serialnumber
        ## result for one spectrometer:
        ## Spectrometers
        ##     - Spectrometer
        ##         - ...
        ## result for 2 spectrometers:
        ## Spectrometers
        ##     - Spectrometer1
        ##         - ...
        ##     - Spectrometer2
        ##         - ...
        for spectrometer in metadata_hardware["Hardware"]["Spectrometers"][
            "Spectrometer"
        ]:
            if spectrometer["Serialnumber"] in spectrometer_serial_numbers:
                idx = spectrometer_serial_numbers.index(spectrometer["Serialnumber"])
                spectrometer_name = spectrometer_serialized_list[idx]
                metadata_hardware["Hardware"]["Spectrometers"][spectrometer_name] = (
                    spectrometer
                )
                ## filter grating via groove density
                gratings_root = spectrometer["Gratings"]["Grating"]
                for grating in gratings_root:
                    if (
                        grating["GrooveDensity"]
                        == infoSerialized_processed["Spectrometers"][spectrometer_name][
                            "Groove_Density"
                        ]
                    ):
                        metadata_hardware["Hardware"]["Spectrometers"][
                            spectrometer_name
                        ]["Gratings"]["Grating"] = grating
        if not spectrometer_name == "Spectrometer":
            del metadata_hardware["Hardware"]["Spectrometers"]["Spectrometer"]

    def _map_general_md(self, original_metadata):
        general = {}
        general["title"] = self._file_path.name.split(".")[0]
        general["original_filename"] = self._file_path.name
        try:
            date, time = original_metadata["Document"]["RecordTime"].split(" ")
        except KeyError:  # pragma: no cover
            pass  # pragma: no cover
        else:
            date_split = date.split("/")
            date = date_split[-1] + "-" + date_split[0] + "-" + date_split[1]
            general["date"] = date
            general["time"] = time.split(".")[0]
        return general

    def _map_signal_md(self, original_metadata):
        signal = {}

        if importlib.util.find_spec("lumispy") is None:
            _logger.warning(
                "Cannot find package lumispy, using BaseSignal as signal_type."
            )
            signal["signal_type"] = ""
        else:
            signal["signal_type"] = "Luminescence"  # pragma: no cover

        try:
            quantity = original_metadata["Document"]["Label"]
            quantity_unit = original_metadata["Document"]["DataLabel"]
        except KeyError:  # pragma: no cover
            pass  # pragma: no cover
        else:
            signal["quantity"] = f"{quantity} ({quantity_unit})"
        return signal

    def _map_detector_md(self, original_metadata):
        detector = {"processing": {}}
        detector_original = original_metadata["Document"]["InfoSerialized"]["Detector"]
        try:
            experiment_original = original_metadata["Document"]["InfoSerialized"][
                "Experiment"
            ]
        except KeyError:
            pass
        else:
            if "Overlap (%)" in experiment_original:
                detector["glued_spectrum"] = True
                detector["glued_spectrum_overlap"] = float(
                    experiment_original.get("Overlap (%)")
                )
                detector["glued_spectrum_windows"] = self._num_datasets
            else:
                detector["glued_spectrum"] = False
        detector["temperature"] = _convert_float(
            detector_original.get("Detector_Temperature")
        )
        detector["exposure_per_frame"] = (
            _convert_float(detector_original.get("Exposure_Time_(ms)")) / 1000
        )
        detector["frames"] = _convert_float(
            detector_original.get("No_of_Accumulations")
        )
        detector["processing"]["calc_average"] = detector_original.get("Calc_Average")

        try:
            detector["exposure_per_frame"] = (
                _convert_float(detector_original.get("Exposure_Time_(ms)")) / 1000
            )
        except TypeError:  # pragma: no cover
            detector["exposure_per_frame"] = None  # pragma: no cover

        if detector["processing"]["calc_average"] == "False":
            try:
                detector["integration_time"] = (
                    detector["exposure_per_frame"] * detector["frames"]
                )
            except TypeError:  # pragma: no cover
                pass  # pragma: no cover
        elif detector["processing"]["calc_average"] == "True":
            detector["integration_time"] = detector["exposure_per_frame"]

        return detector

    @staticmethod
    def _map_laser_md(original_metadata, laser_wavelength):
        laser = {}
        laser["objective_magnification"] = float(
            original_metadata["Hardware"]["Microscopes"]["Microscope"]["Objectives"][
                "Objective"
            ]["Magnification"]
        )
        if laser_wavelength is not None:
            laser["wavelength"] = laser_wavelength
        return laser

    @staticmethod
    def _map_spectrometer_md(original_metadata, central_wavelength):
        all_spectrometers_dict = {}

        spectrometers_original = original_metadata["Document"]["InfoSerialized"][
            "Spectrometers"
        ]

        for key, entry in spectrometers_original.items():
            spectro_dict_tmp = {"Grating": {}}
            spectro_dict_tmp["central_wavelength"] = central_wavelength
            blaze = original_metadata["Hardware"]["Spectrometers"][key]["Gratings"][
                "Grating"
            ]["Blaze"]
            if blaze[-2:] == "NM":
                blaze = float(blaze.split("N")[0])
            spectro_dict_tmp["Grating"]["blazing_wavelength"] = blaze
            spectro_dict_tmp["model"] = entry.get("Model")
            try:
                groove_density = entry["Groove_Density"]
            except KeyError:  # pragma: no cover
                groove_density = None  # pragma: no cover
            else:
                groove_density = float(groove_density.split(" ")[0])
            spectro_dict_tmp["Grating"]["groove_density"] = groove_density
            slit_entrance_front = (
                _convert_float(entry.get("Slit_Entrance-Front")) / 1000
            )
            slit_entrance_side = _convert_float(entry.get("Slit_Entrance-Side")) / 1000
            slit_exit_front = _convert_float(entry.get("Slit_Exit-Front")) / 1000
            slit_exit_side = _convert_float(entry.get("Slit_Exit-Side")) / 1000
            ## using the maximum here, because
            ## only one entrance/exit should be in use anyways
            spectro_dict_tmp["entrance_slit_width"] = max(
                slit_entrance_front, slit_entrance_side
            )
            spectro_dict_tmp["exit_slit_width"] = max(slit_exit_front, slit_exit_side)
            all_spectrometers_dict[key] = spectro_dict_tmp

        return all_spectrometers_dict

    @staticmethod
    def _get_calibration_md(original_metadata):
        try:
            calibration_original = original_metadata["Document"]["InfoSerialized"][
                "Calibration"
            ]
        except KeyError:
            central_wavelength = None
            laser_wavelength = None
        else:
            central_wavelength = _convert_float(
                calibration_original.get("Center_Wavelength")
            )
            laser_wavelength = _convert_float(
                calibration_original.get("Laser_Wavelength")
            )
            if laser_wavelength is not None:
                if np.isclose(laser_wavelength, 0):
                    laser_wavelength = None
        return central_wavelength, laser_wavelength

    def map_metadata(self, original_metadata):
        """Maps original_metadata to metadata."""
        general = self._map_general_md(original_metadata)
        signal = self._map_signal_md(original_metadata)
        detector = self._map_detector_md(original_metadata)
        central_wavelength, laser_wavelength = self._get_calibration_md(
            original_metadata
        )
        laser = self._map_laser_md(original_metadata, laser_wavelength)
        spectrometer = self._map_spectrometer_md(original_metadata, central_wavelength)

        acquisition_instrument = {
            "Detector": detector,
            "Laser": laser,
        }
        acquisition_instrument.update(spectrometer)

        metadata = {
            "Acquisition_instrument": acquisition_instrument,
            "General": general,
            "Signal": signal,
        }
        _remove_none_from_dict(metadata)
        return metadata

    def _get_signal_axis(self, axis_pos):
        """Helper method to read and set signal axis."""
        axis_list = axis_pos.findall("xDim")
        _error_handling_find_location(
            len(axis_list), "signal axis information"
        )  # pragma: no cover
        axis = axis_list[0]
        signal_data = axis[0].attrib["ValueArray"].split("|")
        if len(signal_data) != (int(signal_data[0]) + 1):
            _logger.critical(
                "Signal data size does not match expected size."
            )  # pragma: no cover
        signal_data = np.array([float(x) for x in signal_data[1:]])
        signal_dict = {}
        signal_dict["name"] = "Wavelength"
        unit = axis[0].attrib["Unit"]
        if unit == "Nanometer":
            signal_dict["units"] = "nm"
        else:
            signal_dict["units"] = unit  # pragma: no cover
        signal_dict["navigate"] = False
        if signal_data.size > 1:
            if self._use_uniform_signal_axis:
                offset, scale = polyfit(np.arange(signal_data.size), signal_data, deg=1)
                signal_dict["offset"] = offset
                signal_dict["scale"] = scale
                signal_dict["size"] = signal_data.size
                scale_compare = 100 * np.max(
                    np.abs(np.diff(signal_data) - scale) / scale
                )
                if scale_compare > 1:
                    _logger.warning(
                        f"The relative variation of the signal-axis-scale ({scale_compare:.2f}%) exceeds 1%.\n"
                        "                            "
                        "Using a non-uniform-axis is recommended."
                    )
            else:
                signal_dict["axis"] = signal_data
        else:  # pragma: no cover
            if self._use_uniform_signal_axis:  # pragma: no cover
                _logger.warning(  # pragma: no cover
                    "Signal only contains one entry.\n"  # pragma: no cover
                    "                            "  # pragma: no cover
                    "Using non-uniform-axis independent of use_uniform_signal_axis setting"  # pragma: no cover
                )  # pragma: no cover
            signal_dict["axis"] = signal_data  # pragma: no cover
        return signal_dict

    @staticmethod
    def _get_time_axis(time, axes_dict):
        scale = time[1]
        size = time.size
        ## inconsistency between timestamps and metadata
        ## (Document/InfoSerialized/Experiment6)
        ## in timeseries example file:
        ## exposure time: 1 sec
        ## delay: 3 sec
        ## accumulations: 2
        ## frames: 10
        ## total time: 56 sec
        ## timestamp scale: 4 sec
        ## -> max timestamp: 36 sec < 56 sec
        ## Here the timestamp is used for scale

        axes_dict["time"] = {
            "name": "time",
            "units": "s",
            "size": size,
            "offset": 0,
            "scale": scale,
            "navigate": False,
        }
        axes_dict["time"]["index_in_array"] = len(axes_dict) - 1

    @staticmethod
    def _get_nav_axis(name, axis):
        """Helper method to read and set navigation axes."""
        nav_dict = {}
        nav_dict["offset"] = float(axis["From"])
        nav_dict["scale"] = float(axis["Step"])
        nav_dict["size"] = int(axis["Points"])
        nav_dict["navigate"] = True
        nav_dict["name"] = name
        nav_dict["units"] = "µm"
        return nav_dict

    def set_axes(self, axis_head, signal_axis, time):
        """Extracts signal and navigation axes."""
        axes = dict()
        has_y = False
        has_x = False
        num_xy_frames = 0
        if "Y-Axis" in axis_head.keys():
            axes["Y"] = self._get_nav_axis("Y", axis_head["Y-Axis"])
            num_xy_frames += axes["Y"]["size"]
            axes["Y"]["index_in_array"] = 0
            has_y = True
        if "X-Axis" in axis_head.keys():
            axes["X"] = self._get_nav_axis("X", axis_head["X-Axis"])
            has_x = True
            if has_y:
                num_xy_frames *= axes["X"]["size"]
                axes["X"]["index_in_array"] = 1
            else:
                num_xy_frames += axes["X"]["size"]
                axes["X"]["index_in_array"] = 0

        ## adding time-axis entry if appropriate
        ## this is done inplace (the argument "axes" itself is altered)
        total_frames = int(axis_head["Detector"]["No_of_Frames"])
        if (total_frames - num_xy_frames) == 0 or total_frames == 1:
            has_time = False
        else:
            has_time = True
            self._get_time_axis(time, axes)
        if (has_x or has_y) and has_time:
            raise NotImplementedError(  # pragma: no cover
                "Reading a combination of timeseries and map or linescan is not implemented."
            )

        if self._glued_data_as_stack and self._num_datasets != 0:
            if has_time or has_x or has_y:
                _logger.warning(  # pragma: no cover
                    "Loading glued data as stack in combination with multiple axis (time, linescan or map)"
                    "                            "
                    "is not tested and may lead to false results."
                    "                            "
                    "Please use glued_data_as_stack=False for loading the file."
                )

            axes_list = []
            for signal_axis in signal_axis:
                axes_tmp = deepcopy(axes)
                axes_tmp["signal_dict"] = signal_axis
                axes_tmp["signal_dict"]["index_in_array"] = len(axes) - 1
                axes_tmp = sorted(
                    axes_tmp.values(), key=lambda item: item["index_in_array"]
                )
                axes_list.append(axes_tmp)
        else:
            axes["signal_dict"] = signal_axis[0]
            axes["signal_dict"]["index_in_array"] = len(axes) - 1
            ## extra surrounding list here to ensure compatibility
            ## with glued_data_as_stack for file_reader()
            axes_list = [sorted(axes.values(), key=lambda item: item["index_in_array"])]

        return axes_list

    @staticmethod
    def _parse_data(data_pos):
        """Extracts data from file."""
        data_list = data_pos.findall("Data")
        _error_handling_find_location(len(data_list), "data")  # pragma: no cover

        ## dtype=np.int64 instead of int here,
        ## because on windows python int defaults to 32bit
        ## the timestamp is given as windows filetime
        ## -> number is too large for 32bit
        data_array = []
        time_array = []
        for frame in data_list[0]:
            time_frame = np.fromstring(
                frame.attrib["TimeStamp"], sep=" ", dtype=np.int64
            )
            data_frame = [float(x) for x in frame.text.split(";")]
            time_array.append(time_frame)
            data_array.append(data_frame)
        data = np.array(data_array).ravel()
        time = (np.array(time_array, dtype=np.int64) - time_array[0]).ravel() / 1e7
        return data, time

    def _load_glued_data_stack(self, data_head):
        num_datasets_list = data_head.findall("Childs")  # codespell:ignore
        _error_handling_find_location(
            len(num_datasets_list), "glued datasets"
        )  # pragma: no cover
        data_array = []
        signal_axis_list = []
        time_array = []
        for dataset in num_datasets_list[0]:
            signal_axis = self._get_signal_axis(dataset)
            signal_axis_list.append(signal_axis)
            data, time = self._parse_data(dataset)
            data_array.append(data)
            time_array.append(time)
        data = np.array(data_array)
        time = np.array(time_array)
        return data, time, signal_axis_list

    def get_data_and_signal(self, data_head):
        if self._glued_data_as_stack and self._num_datasets != 0:
            data, time, signal_axis = self._load_glued_data_stack(data_head)
        else:
            data, time = self._parse_data(data_head)
            signal_axis = [self._get_signal_axis(data_head)]
            data = [data]
            ## extra surrounding list here
            ## to ensure compatibility with glued_data_as_stack=True
            ## for file_reader(), reshape_data()
        return data, time, signal_axis

    def reshape_data(self, data, axes):
        """Reshapes data according to axes sizes."""
        if self._use_uniform_signal_axis:
            wavelength_size = axes[0][-1]["size"]
        else:
            wavelength_size = axes[0][-1]["axis"].size
        shape_sizes = []
        for i in range(len(axes[0]) - 1):
            shape_sizes.append(axes[0][i]["size"])
        shape_sizes.append(wavelength_size)

        for i, dataset in enumerate(data):
            dataset_reshaped = np.reshape(dataset, shape_sizes)
            data[i] = dataset_reshaped
        return data


def file_reader(
    filename,
    lazy=False,
    use_uniform_signal_axis=False,
    glued_data_as_stack=False,
    filter_original_metadata=True,
):
    """
    Read TriVista's ``.tvf`` file.

    Parameters
    ----------
    %s
    %s
    use_uniform_signal_axis : bool, default=False
        Can be specified to choose between non-uniform or uniform signal axes.
        If `True`, the ``scale`` attribute is calculated from the average delta
        along the signal axis and a warning is raised in case the delta varies
        by more than 1%%.
    glued_data_as_stack : bool, default=False
        Using the mode `Step & Glue` results in measurements performed
        at different wavelength ranges with some overlap between them.
        The file then contains the individual spectra as well as
        the "glued" spectrum. The latter is represented as one spectrum,
        which covers the complete wavelength range. Stitching the datasets
        together in the overlap region is already done by the setup.
        If this setting is set to `True`, then the individual datasets will be loaded
        as a stack. Otherwise, only the "glued" spectrum is loaded.
    filter_original_metadata : bool, default=True
        Decides whether to process the original_metadata.
        If `True`, then non-relevant metadata will be excluded.
        For example, the metadata usually contains information
        for multiple objectives, even though only one is used.
        In this case, only the metadata from the used objective
        will be added to original_metadata.
        This setting only affects the ``original_metadata`` attribute
        and not the ``metadata`` attribute.

    %s
    """
    if lazy is not False:
        raise NotImplementedError("Lazy loading is not supported.")

    t = TrivistaTVFReader(
        Path(filename),
        use_uniform_signal_axis=use_uniform_signal_axis,
        glued_data_as_stack=glued_data_as_stack,
        filter_original_metadata=filter_original_metadata,
    )

    result = []
    for dataset, axes in zip(t.data, t.axes):
        result.append(
            {
                "data": dataset,
                "axes": axes,
                "metadata": deepcopy(t.metadata),
                "original_metadata": deepcopy(t.original_metadata),
            }
        )
    return result


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)
